@@ -33,6 +33,8 @@ extern char *optarg;
 #include "uuid/uuid.h"
 #include "e2p/e2p.h"
 
+#include <ext2fs/ext2_ext_attr.h>
+
 #include "../version.h"
 
 extern ss_request_table debug_cmds;
@@ -41,9 +43,11 @@ ext2_filsys	current_fs = NULL;
 ext2_ino_t	root, cwd;
 
 static void open_filesystem(char *device, int open_flags, blk_t superblock,
-			    blk_t blocksize, int catastrophic)
+			    blk_t blocksize, int catastrophic, 
+			    char *data_filename)
 {
 	int	retval;
+	io_channel data_io = 0;
 
 	if (superblock != 0 && blocksize == 0) {
 		com_err(device, 0, "if you specify the superblock, you must also specify the block size");
@@ -51,6 +55,21 @@ static void open_filesystem(char *device, int open_flags, blk_t superblock,
 		return;
 	}
 
+	if (data_filename) {
+		if ((open_flags & EXT2_FLAG_IMAGE_FILE) == 0) {
+			com_err(device, 0, 
+				"The -d option is only valid when reading an e2image file");
+			current_fs = NULL;
+			return;
+		}
+		retval = unix_io_manager->open(data_filename, 0, &data_io);
+		if (retval) {
+			com_err(data_filename, 0, "while opening data source");
+			current_fs = NULL;
+			return;
+		}
+	}
+
 	if (catastrophic && (open_flags & EXT2_FLAG_RW)) {
 		com_err(device, 0,
 			"opening read-only because of catastrophic mode");
@@ -79,6 +98,16 @@ static void open_filesystem(char *device, int open_flags, blk_t superblock,
 			goto errout;
 		}
 	}
+
+	if (data_io) {
+		retval = ext2fs_set_data_io(current_fs, data_io);
+		if (retval) {
+			com_err(device, retval, 
+				"while setting data source");
+			goto errout;
+		}
+	}
+
 	root = cwd = EXT2_ROOT_INO;
 	return;
 
@@ -96,10 +125,11 @@ void do_open_filesys(int argc, char **argv)
 	int	catastrophic = 0;
 	blk_t	superblock = 0;
 	blk_t	blocksize = 0;
-	int open_flags = 0;
+	int	open_flags = 0;
+	char	*data_filename = 0;
 	
 	reset_getopt();
-	while ((c = getopt (argc, argv, "iwfcb:s:")) != EOF) {
+	while ((c = getopt (argc, argv, "iwfcb:s:d:")) != EOF) {
 		switch (c) {
 		case 'i':
 			open_flags |= EXT2_FLAG_IMAGE_FILE;
@@ -113,6 +143,9 @@ void do_open_filesys(int argc, char **argv)
 		case 'c':
 			catastrophic = 1;
 			break;
+		case 'd':
+			data_filename = optarg;
+			break;
 		case 'b':
 			blocksize = parse_ulong(optarg, argv[0],
 						"block size", &err);
@@ -137,7 +170,8 @@ void do_open_filesys(int argc, char **argv)
 	if (check_fs_not_open(argv[0]))
 		return;
 	open_filesystem(argv[optind], open_flags,
-			superblock, blocksize, catastrophic);
+			superblock, blocksize, catastrophic, 
+			data_filename);
 }
 
 void do_lcd(int argc, char **argv)
@@ -229,7 +263,7 @@ static void print_features(struct ext2_super_block * s, FILE *f)
 
 void do_show_super_stats(int argc, char *argv[])
 {
-	int	i;
+	dgrp_t	i;
 	FILE 	*out;
 	struct ext2_group_desc *gdp;
 	int	c, header_only = 0;
@@ -285,7 +319,8 @@ void do_show_super_stats(int argc, char *argv[])
 	close_pager(out);
 }
 
-void do_dirty_filesys(int argc, char **argv)
+void do_dirty_filesys(int argc EXT2FS_ATTR((unused)), 
+		      char **argv EXT2FS_ATTR((unused)))
 {
 	if (check_fs_open(argv[0]))
 		return;
@@ -323,9 +358,11 @@ static void finish_range(struct list_blocks_struct *lb)
 	lb->first_block = 0;
 }
 
-static int list_blocks_proc(ext2_filsys fs, blk_t *blocknr,
-			    e2_blkcnt_t blockcnt, blk_t ref_block,
-			    int ref_offset, void *private)
+static int list_blocks_proc(ext2_filsys fs EXT2FS_ATTR((unused)), 
+			    blk_t *blocknr, e2_blkcnt_t blockcnt, 
+			    blk_t ref_block EXT2FS_ATTR((unused)),
+			    int ref_offset EXT2FS_ATTR((unused)), 
+			    void *private)
 {
 	struct list_blocks_struct *lb = (struct list_blocks_struct *) private;
 
@@ -366,6 +403,70 @@ static int list_blocks_proc(ext2_filsys fs, blk_t *blocknr,
 	return 0;
 }
 
+void dump_xattr_string(FILE *out, const unsigned char *str, int len)
+{
+	int printable = 1;
+	int i;
+	
+	/* check is string printable? */
+	for (i = 0; i < len; i++)
+		if (!isprint(str[i])) {
+			printable = 0;
+			break;
+		}
+
+	for (i = 0; i < len; i++)
+		if (printable)
+			fprintf(out, "%c", str[i]);
+		else
+			fprintf(out, "%02x ", str[i]);
+}
+
+void internal_dump_inode_extra(FILE *out, const char *prefix,
+			 ext2_ino_t inode_num, struct ext2_inode_large *inode)
+{
+	struct ext2_ext_attr_entry *entry;
+	__u32 *magic;
+	char *start, *end;
+	int storage_size;
+	int i;
+
+	fprintf(out, "Size of extra inode fields: %d\n", inode->i_extra_isize);
+	if (inode->i_extra_isize > EXT2_INODE_SIZE(current_fs->super) -
+			EXT2_GOOD_OLD_INODE_SIZE) {
+		fprintf(stderr, "invalid inode->i_extra_isize (%u)\n",
+				inode->i_extra_isize);
+		return;
+	}
+	storage_size = EXT2_INODE_SIZE(current_fs->super) -
+			EXT2_GOOD_OLD_INODE_SIZE -
+			inode->i_extra_isize;
+	magic = (__u32 *)((char *)inode + EXT2_GOOD_OLD_INODE_SIZE +
+			inode->i_extra_isize);
+	if (*magic == EXT2_EXT_ATTR_MAGIC) {
+		fprintf(out, "Extended attributes stored in inode body: \n");
+		end = (char *) inode + EXT2_INODE_SIZE(current_fs->super);
+		start = (char *) magic + sizeof(__u32);
+		entry = (struct ext2_ext_attr_entry *) start;
+		while (!EXT2_EXT_IS_LAST_ENTRY(entry)) {
+			struct ext2_ext_attr_entry *next =
+				EXT2_EXT_ATTR_NEXT(entry);
+			if (entry->e_value_size > storage_size ||
+					(char *) next >= end) {
+				fprintf(out, "invalid EA entry in inode\n");
+				return;
+			}
+			fprintf(out, "  ");
+			dump_xattr_string(out, EXT2_EXT_ATTR_NAME(entry), 
+					  entry->e_name_len);
+			fprintf(out, " = \"");
+			dump_xattr_string(out, start + entry->e_value_offs,
+						entry->e_value_size);
+			fprintf(out, "\" (%d)\n", entry->e_value_size);
+			entry = next;
+		}
+	}
+}
 
 static void dump_blocks(FILE *f, const char *prefix, ext2_ino_t inode)
 {
@@ -453,34 +554,67 @@ void internal_dump_inode(FILE *out, const char *prefix,
 	if (inode->i_dtime) 
 	  fprintf(out, "%sdtime: 0x%08x -- %s", prefix, inode->i_dtime,
 		  time_to_string(inode->i_dtime));
-	if (LINUX_S_ISLNK(inode->i_mode) && inode->i_blocks == 0)
+	if (EXT2_INODE_SIZE(current_fs->super) > EXT2_GOOD_OLD_INODE_SIZE)
+		internal_dump_inode_extra(out, prefix, inode_num,
+					  (struct ext2_inode_large *) inode);
+	if (LINUX_S_ISLNK(inode->i_mode) && ext2fs_inode_data_blocks(current_fs,inode) == 0)
 		fprintf(out, "%sFast_link_dest: %.*s\n", prefix,
 			(int) inode->i_size, (char *)inode->i_block);
+	else if (LINUX_S_ISBLK(inode->i_mode) || LINUX_S_ISCHR(inode->i_mode)) {
+		int major, minor;
+		const char *devnote;
+
+		if (inode->i_block[0]) {
+			major = (inode->i_block[0] >> 8) & 255;
+			minor = inode->i_block[0] & 255;
+			devnote = "";
+		} else {
+			major = (inode->i_block[1] & 0xfff00) >> 8;
+			minor = ((inode->i_block[1] & 0xff) | 
+				 ((inode->i_block[1] >> 12) & 0xfff00));
+			devnote = "(New-style) ";
+		}
+		fprintf(out, "%sDevice major/minor number: %02d:%02d (hex %02x:%02x)\n", 
+			devnote, major, minor, major, minor);
+	}
 	else if (do_dump_blocks)
 		dump_blocks(out, prefix, inode_num);
 }
 
-static void dump_inode(ext2_ino_t inode_num, struct ext2_inode inode)
+static void dump_inode(ext2_ino_t inode_num, struct ext2_inode *inode)
 {
 	FILE	*out;
 	
 	out = open_pager();
-	internal_dump_inode(out, "", inode_num, &inode, 1);
+	internal_dump_inode(out, "", inode_num, inode, 1);
 	close_pager(out);
 }
 
 void do_stat(int argc, char *argv[])
 {
 	ext2_ino_t	inode;
-	struct ext2_inode inode_buf;
+	struct ext2_inode * inode_buf;
 
-	if (common_inode_args_process(argc, argv, &inode, 0))
+	inode_buf = (struct ext2_inode *)
+			malloc(EXT2_INODE_SIZE(current_fs->super));
+	if (!inode_buf) {
+		fprintf(stderr, "do_stat: can't allocate buffer\n");
 		return;
+	}
 
-	if (debugfs_read_inode(inode, &inode_buf, argv[0]))
+	if (common_inode_args_process(argc, argv, &inode, 0)) {
+		free(inode_buf);
 		return;
+	}
 
-	dump_inode(inode,inode_buf);
+	if (debugfs_read_inode_full(inode, inode_buf, argv[0],
+					EXT2_INODE_SIZE(current_fs->super))) {
+		free(inode_buf);
+		return;
+	}
+
+	dump_inode(inode, inode_buf);
+	free(inode_buf);
 	return;
 }
 
@@ -798,9 +932,39 @@ void do_print_working_directory(int argc, char *argv[])
 	return;
 }
 
+/*
+ * Given a mode, return the ext2 file type
+ */
+static int ext2_file_type(unsigned int mode)
+{
+	if (LINUX_S_ISREG(mode))
+		return EXT2_FT_REG_FILE;
+
+	if (LINUX_S_ISDIR(mode))
+		return EXT2_FT_DIR;
+	
+	if (LINUX_S_ISCHR(mode))
+		return EXT2_FT_CHRDEV;
+	
+	if (LINUX_S_ISBLK(mode))
+		return EXT2_FT_BLKDEV;
+	
+	if (LINUX_S_ISLNK(mode))
+		return EXT2_FT_SYMLINK;
+
+	if (LINUX_S_ISFIFO(mode))
+		return EXT2_FT_FIFO;
+	
+	if (LINUX_S_ISSOCK(mode))
+		return EXT2_FT_SOCK;
+	
+	return 0;
+}
+
 static void make_link(char *sourcename, char *destname)
 {
-	ext2_ino_t	inode;
+	ext2_ino_t	ino;
+	struct ext2_inode inode;
 	int		retval;
 	ext2_ino_t	dir;
 	char		*dest, *cp, *basename;
@@ -808,8 +972,8 @@ static void make_link(char *sourcename, char *destname)
 	/*
 	 * Get the source inode
 	 */
-	inode = string_to_inode(sourcename);
-	if (!inode)
+	ino = string_to_inode(sourcename);
+	if (!ino)
 		return;
 	basename = strrchr(sourcename, '/');
 	if (basename)
@@ -839,8 +1003,12 @@ static void make_link(char *sourcename, char *destname)
 			dest = destname;
 		}
 	}
+
+	if (debugfs_read_inode(ino, &inode, sourcename))
+		return;
 	
-	retval = ext2fs_link(current_fs, dir, dest, inode, 0);
+	retval = ext2fs_link(current_fs, dir, dest, ino, 
+			     ext2_file_type(inode.i_mode));
 	if (retval)
 		com_err("make_link", retval, "");
 	return;
@@ -857,7 +1025,8 @@ void do_link(int argc, char *argv[])
 }
 
 static int mark_blocks_proc(ext2_filsys fs, blk_t *blocknr,
-			    int blockcnt, void *private)
+			    int blockcnt EXT2FS_ATTR((unused)), 
+			    void *private EXT2FS_ATTR((unused)))
 {
 	blk_t	block;
 
@@ -1104,6 +1273,15 @@ void do_write(int argc, char *argv[])
 	printf("Allocated inode: %u\n", newfile);
 	retval = ext2fs_link(current_fs, cwd, argv[2], newfile,
 			     EXT2_FT_REG_FILE);
+	if (retval == EXT2_ET_DIR_NO_SPACE) {
+		retval = ext2fs_expand_dir(current_fs, cwd);
+		if (retval) {
+			com_err(argv[0], retval, "while expanding directory");
+			return;
+		}
+		retval = ext2fs_link(current_fs, cwd, argv[2], newfile,
+				     EXT2_FT_REG_FILE);
+	}
 	if (retval) {
 		com_err(argv[2], retval, "");
 		close(fd);
@@ -1113,11 +1291,11 @@ void do_write(int argc, char *argv[])
 		com_err(argv[0], 0, "Warning: inode already set");
 	ext2fs_inode_alloc_stats2(current_fs, newfile, +1, 0);
 	memset(&inode, 0, sizeof(inode));
-	inode.i_mode = statbuf.st_mode;
+	inode.i_mode = (statbuf.st_mode & ~LINUX_S_IFMT) | LINUX_S_IFREG;
 	inode.i_atime = inode.i_ctime = inode.i_mtime = time(NULL);
 	inode.i_links_count = 1;
 	inode.i_size = statbuf.st_size;
-	if (debugfs_write_inode(newfile, &inode, argv[0])) {
+	if (debugfs_write_new_inode(newfile, &inode, argv[0])) {
 		close(fd);
 		return;
 	}
@@ -1131,11 +1309,11 @@ void do_write(int argc, char *argv[])
 
 void do_mknod(int argc, char *argv[])
 {
-	unsigned long	mode, major, minor, nr;
+	unsigned long	mode, major, minor;
 	ext2_ino_t	newfile;
 	errcode_t 	retval;
 	struct ext2_inode inode;
-	int		filetype;
+	int		filetype, nr;
 
 	if (check_fs_open(argv[0]))
 		return;
@@ -1168,7 +1346,7 @@ void do_mknod(int argc, char *argv[])
 	if (nr == 5) {
 		major = strtoul(argv[3], argv+3, 0);
 		minor = strtoul(argv[4], argv+4, 0);
-		if (major > 255 || minor > 255 || argv[3][0] || argv[4][0])
+		if (major > 65535 || minor > 65535 || argv[3][0] || argv[4][0])
 			nr = 0;
 	}
 	if (argc != nr)
@@ -1182,18 +1360,18 @@ void do_mknod(int argc, char *argv[])
 	}
 	printf("Allocated inode: %u\n", newfile);
 	retval = ext2fs_link(current_fs, cwd, argv[1], newfile, filetype);
-	if (retval) {
-		if (retval == EXT2_ET_DIR_NO_SPACE) {
-			retval = ext2fs_expand_dir(current_fs, cwd);
-			if (!retval)
-				retval = ext2fs_link(current_fs, cwd,
-						     argv[1], newfile,
-						     filetype);
-		}
+	if (retval == EXT2_ET_DIR_NO_SPACE) {
+		retval = ext2fs_expand_dir(current_fs, cwd);
 		if (retval) {
-			com_err(argv[1], retval, "");
+			com_err(argv[0], retval, "while expanding directory");
 			return;
 		}
+		retval = ext2fs_link(current_fs, cwd, argv[1], newfile,
+				     filetype);
+	}
+	if (retval) {
+		com_err(argv[1], retval, "");
+		return;
 	}
         if (ext2fs_test_inode_bitmap(current_fs->inode_map,newfile))
 		com_err(argv[0], 0, "Warning: inode already set");
@@ -1202,9 +1380,15 @@ void do_mknod(int argc, char *argv[])
 	memset(&inode, 0, sizeof(inode));
 	inode.i_mode = mode;
 	inode.i_atime = inode.i_ctime = inode.i_mtime = time(NULL);
-	inode.i_block[0] = major*256+minor;
+	if ((major < 256) && (minor < 256)) {
+		inode.i_block[0] = major*256+minor;
+		inode.i_block[1] = 0;
+	} else {
+		inode.i_block[0] = 0;
+		inode.i_block[1] = (minor & 0xff) | (major << 8) | ((minor & ~0xff) << 12);
+	}
 	inode.i_links_count = 1;
-	if (debugfs_write_inode(newfile, &inode, argv[0]))
+	if (debugfs_write_new_inode(newfile, &inode, argv[0]))
 		return;
 }
 
@@ -1233,8 +1417,16 @@ void do_mkdir(int argc, char *argv[])
 		name = argv[1];
 	}
 
-
+try_again:
 	retval = ext2fs_mkdir(current_fs, parent, 0, name);
+	if (retval == EXT2_ET_DIR_NO_SPACE) {
+		retval = ext2fs_expand_dir(current_fs, parent);
+		if (retval) {
+			com_err("argv[0]", retval, "while expanding directory");
+			return;
+		}
+		goto try_again;
+	}
 	if (retval) {
 		com_err("ext2fs_mkdir", retval, "");
 		return;
@@ -1243,7 +1435,8 @@ void do_mkdir(int argc, char *argv[])
 }
 
 static int release_blocks_proc(ext2_filsys fs, blk_t *blocknr,
-			       int blockcnt, void *private)
+			       int blockcnt EXT2FS_ATTR((unused)), 
+			       void *private EXT2FS_ATTR((unused)))
 {
 	blk_t	block;
 
@@ -1261,6 +1454,8 @@ static void kill_file_by_inode(ext2_ino_t inode)
 	inode_buf.i_dtime = time(NULL);
 	if (debugfs_write_inode(inode, &inode_buf, 0))
 		return;
+	if (!ext2fs_inode_has_valid_blocks(&inode_buf))
+		return;
 
 	ext2fs_block_iterate(current_fs, inode, 0, NULL,
 			     release_blocks_proc, NULL);
@@ -1318,12 +1513,12 @@ struct rd_struct {
 	int		empty;
 };
 
-static int rmdir_proc(ext2_ino_t dir,
-		      int	entry,
+static int rmdir_proc(ext2_ino_t dir EXT2FS_ATTR((unused)),
+		      int	entry EXT2FS_ATTR((unused)),
 		      struct ext2_dir_entry *dirent,
-		      int	offset,
-		      int	blocksize,
-		      char	*buf,
+		      int	offset EXT2FS_ATTR((unused)),
+		      int	blocksize EXT2FS_ATTR((unused)),
+		      char	*buf EXT2FS_ATTR((unused)),
 		      void	*private)
 {
 	struct rd_struct *rds = (struct rd_struct *) private;
@@ -1397,7 +1592,8 @@ void do_rmdir(int argc, char *argv[])
 	}
 }
 
-void do_show_debugfs_params(int argc, char *argv[])
+void do_show_debugfs_params(int argc EXT2FS_ATTR((unused)), 
+			    char *argv[] EXT2FS_ATTR((unused)))
 {
 	FILE *out = stdout;
 
@@ -1484,7 +1680,7 @@ void do_imap(int argc, char *argv[])
 		EXT2_INODE_SIZE(current_fs->super);
 	block = offset >> EXT2_BLOCK_SIZE_BITS(current_fs->super);
 	if (!current_fs->group_desc[(unsigned)group].bg_inode_table) {
-		com_err(argv[0], 0, "Inode table for group %d is missing\n",
+		com_err(argv[0], 0, "Inode table for group %lu is missing\n",
 			group);
 		return;
 	}
@@ -1551,12 +1747,13 @@ int main(int argc, char **argv)
 	blk_t		superblock = 0;
 	blk_t		blocksize = 0;
 	int		catastrophic = 0;
+	char		*data_filename = 0;
 	
 	initialize_ext2_error_table();
 	fprintf (stderr, "debugfs %s (%s)\n", E2FSPROGS_VERSION,
 		 E2FSPROGS_DATE);
 
-	while ((c = getopt (argc, argv, "iwcR:f:b:s:V")) != EOF) {
+	while ((c = getopt (argc, argv, "iwcR:f:b:s:Vd:")) != EOF) {
 		switch (c) {
 		case 'R':
 			request = optarg;
@@ -1564,6 +1761,9 @@ int main(int argc, char **argv)
 		case 'f':
 			cmd_file = optarg;
 			break;
+		case 'd':
+			data_filename = optarg;
+			break;
 		case 'i':
 			open_flags |= EXT2_FLAG_IMAGE_FILE;
 			break;
@@ -1593,7 +1793,8 @@ int main(int argc, char **argv)
 	}
 	if (optind < argc)
 		open_filesystem(argv[optind], open_flags,
-				superblock, blocksize, catastrophic);
+				superblock, blocksize, catastrophic,
+				data_filename);
 	
 	sci_idx = ss_create_invocation("debugfs", "0.0", (char *) NULL,
 				       &debug_cmds, &retval);
