@@ -41,6 +41,7 @@
  * 	- The inode_reg_map bitmap
  */
 
+#define _GNU_SOURCE 1 /* get strnlen() */
 #include <string.h>
 
 #include "e2fsck.h"
@@ -96,7 +97,7 @@ void e2fsck_pass2(e2fsck_t ctx)
 	struct check_dir_struct cd;
 	struct dx_dir_info	*dx_dir;
 	struct dx_dirblock_info	*dx_db, *dx_parent;
-	blk_t			b;
+	int			b;
 	int			i, depth;
 	problem_t		code;
 	int			bad_dir;
@@ -242,16 +243,9 @@ void e2fsck_pass2(e2fsck_t ctx)
 			clear_htree(ctx, dx_dir->ino);
 			dx_dir->numblocks = 0;
 		}
-#ifdef ENABLE_HTREE_CLEAR
-		if (dx_dir->numblocks) {
-			fix_problem(ctx, PR_2_HTREE_FCLR, &pctx);
-			clear_htree(ctx, dx_dir->ino);
-			dx_dir->numblocks = 0;
-		}
-#endif
 	}
 #endif
-	ext2fs_free_mem((void **) &buf);
+	ext2fs_free_mem(&buf);
 	ext2fs_free_dblist(fs->dblist);
 
 	if (ctx->inode_bad_map) {
@@ -459,7 +453,8 @@ static int check_dotdot(e2fsck_t ctx,
  */
 static int check_name(e2fsck_t ctx,
 		      struct ext2_dir_entry *dirent,
-		      ext2_ino_t dir_ino, struct problem_context *pctx)
+		      ext2_ino_t dir_ino EXT2FS_ATTR((unused)), 
+		      struct problem_context *pctx)
 {
 	int	i;
 	int	fixup = -1;
@@ -483,8 +478,9 @@ static int check_name(e2fsck_t ctx,
  * Check the directory filetype (if present)
  */
 static _INLINE_ int check_filetype(e2fsck_t ctx,
-		      struct ext2_dir_entry *dirent,
-		      ext2_ino_t dir_ino, struct problem_context *pctx)
+				   struct ext2_dir_entry *dirent,
+				   ext2_ino_t dir_ino EXT2FS_ATTR((unused)),
+				   struct problem_context *pctx)
 {
 	int	filetype = dirent->name_len >> 8;
 	int	should_be = EXT2_FT_UNKNOWN;
@@ -591,7 +587,7 @@ static void parse_int_node(ext2_filsys fs,
 #endif
 		blk = ext2fs_le32_to_cpu(ent[i].block) & 0x0ffffff;
 		/* Check to make sure the block is valid */
-		if (blk > dx_dir->numblocks) {
+		if (blk > (blk_t) dx_dir->numblocks) {
 			cd->pctx.blk = blk;
 			if (fix_problem(cd->ctx, PR_2_HTREE_BADBLK,
 					&cd->pctx))
@@ -644,7 +640,7 @@ clear_and_exit:
 static void salvage_directory(ext2_filsys fs,
 			      struct ext2_dir_entry *dirent,
 			      struct ext2_dir_entry *prev,
-			      int *offset)
+			      unsigned int *offset)
 {
 	char	*cp = (char *) dirent;
 	int left = fs->blocksize - *offset - dirent->rec_len;
@@ -708,7 +704,7 @@ static int check_dir_block(ext2_filsys fs,
 #endif /* ENABLE_HTREE */
 	struct ext2_dir_entry 	*dirent, *prev;
 	ext2_dirhash_t		hash;
-	int			offset = 0;
+	unsigned int		offset = 0;
 	int			dir_modified = 0;
 	int			dot_state;
 	blk_t			block_nr = db->blk;
@@ -1071,15 +1067,18 @@ abort_free_dict:
  */
 static int deallocate_inode_block(ext2_filsys fs,
 				  blk_t	*block_nr,
-				  e2_blkcnt_t blockcnt,
-				  blk_t ref_block,
-				  int ref_offset, 
+				  e2_blkcnt_t blockcnt EXT2FS_ATTR((unused)),
+				  blk_t ref_block EXT2FS_ATTR((unused)),
+				  int ref_offset EXT2FS_ATTR((unused)),
 				  void *priv_data)
 {
 	e2fsck_t	ctx = (e2fsck_t) priv_data;
 	
 	if (HOLE_BLKADDR(*block_nr))
 		return 0;
+	if ((*block_nr < fs->super->s_first_data_block) ||
+	    (*block_nr >= fs->super->s_blocks_count))
+		return 0;
 	ext2fs_unmark_block_bitmap(ctx->block_found_map, *block_nr);
 	ext2fs_block_alloc_stats(fs, *block_nr, -1);
 	return 0;
@@ -1172,6 +1171,7 @@ extern int e2fsck_process_bad_inode(e2fsck_t ctx, ext2_ino_t dir,
 	ext2_filsys fs = ctx->fs;
 	struct ext2_inode	inode;
 	int			inode_modified = 0;
+	int			not_fixed = 0;
 	unsigned char		*frag, *fsize;
 	struct problem_context	pctx;
 	int	problem = 0;
@@ -1183,6 +1183,29 @@ extern int e2fsck_process_bad_inode(e2fsck_t ctx, ext2_ino_t dir,
 	pctx.dir = dir;
 	pctx.inode = &inode;
 
+	if (inode.i_file_acl &&
+	    !(fs->super->s_feature_compat & EXT2_FEATURE_COMPAT_EXT_ATTR) &&
+	    fix_problem(ctx, PR_2_FILE_ACL_ZERO, &pctx)) {
+		inode.i_file_acl = 0;
+#ifdef EXT2FS_ENABLE_SWAPFS
+		/* 
+		 * This is a special kludge to deal with long symlinks
+		 * on big endian systems.  i_blocks had already been
+		 * decremented earlier in pass 1, but since i_file_acl
+		 * hadn't yet been cleared, ext2fs_read_inode()
+		 * assumed that the file was short symlink and would
+		 * not have byte swapped i_block[0].  Hence, we have
+		 * to byte-swap it here.
+		 */
+		if (LINUX_S_ISLNK(inode.i_mode) &&
+		    (fs->flags & EXT2_FLAG_SWAP_BYTES) &&
+		    (inode.i_blocks == fs->blocksize >> 9))
+			inode.i_block[0] = ext2fs_swab32(inode.i_block[0]);
+#endif
+		inode_modified++;
+	} else
+		not_fixed++;
+
 	if (!LINUX_S_ISDIR(inode.i_mode) && !LINUX_S_ISREG(inode.i_mode) &&
 	    !LINUX_S_ISCHR(inode.i_mode) && !LINUX_S_ISBLK(inode.i_mode) &&
 	    !LINUX_S_ISLNK(inode.i_mode) && !LINUX_S_ISFIFO(inode.i_mode) &&
@@ -1211,14 +1234,17 @@ extern int e2fsck_process_bad_inode(e2fsck_t ctx, ext2_ino_t dir,
 			if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
 				return 0;
 			return 1;
-		}
+		} else
+			not_fixed++;
 		problem = 0;
 	}
 		
-	if (inode.i_faddr &&
-	    fix_problem(ctx, PR_2_FADDR_ZERO, &pctx)) {
-		inode.i_faddr = 0;
-		inode_modified++;
+	if (inode.i_faddr) {
+		if (fix_problem(ctx, PR_2_FADDR_ZERO, &pctx)) {
+			inode.i_faddr = 0;
+			inode_modified++;
+		} else
+			not_fixed++;
 	}
 
 	switch (fs->super->s_creator_os) {
@@ -1242,7 +1268,8 @@ extern int e2fsck_process_bad_inode(e2fsck_t ctx, ext2_ino_t dir,
 		if (fix_problem(ctx, PR_2_FRAG_ZERO, &pctx)) {
 			*frag = 0;
 			inode_modified++;
-		}
+		} else
+			not_fixed++;
 		pctx.num = 0;
 	}
 	if (fsize && *fsize) {
@@ -1250,31 +1277,33 @@ extern int e2fsck_process_bad_inode(e2fsck_t ctx, ext2_ino_t dir,
 		if (fix_problem(ctx, PR_2_FSIZE_ZERO, &pctx)) {
 			*fsize = 0;
 			inode_modified++;
-		}
+		} else
+			not_fixed++;
 		pctx.num = 0;
 	}
 
-	if (inode.i_file_acl &&
-	    !(fs->super->s_feature_compat & EXT2_FEATURE_COMPAT_EXT_ATTR) &&
-	    fix_problem(ctx, PR_2_FILE_ACL_ZERO, &pctx)) {
-		inode.i_file_acl = 0;
-		inode_modified++;
-	}
 	if (inode.i_file_acl &&
 	    ((inode.i_file_acl < fs->super->s_first_data_block) ||
-	     (inode.i_file_acl >= fs->super->s_blocks_count)) &&
-	    fix_problem(ctx, PR_2_FILE_ACL_BAD, &pctx)) {
-		inode.i_file_acl = 0;
-		inode_modified++;
+	     (inode.i_file_acl >= fs->super->s_blocks_count))) {
+		if (fix_problem(ctx, PR_2_FILE_ACL_BAD, &pctx)) {
+			inode.i_file_acl = 0;
+			inode_modified++;
+		} else
+			not_fixed++;
 	}
 	if (inode.i_dir_acl &&
-	    LINUX_S_ISDIR(inode.i_mode) &&
-	    fix_problem(ctx, PR_2_DIR_ACL_ZERO, &pctx)) {
-		inode.i_dir_acl = 0;
-		inode_modified++;
+	    LINUX_S_ISDIR(inode.i_mode)) {
+		if (fix_problem(ctx, PR_2_DIR_ACL_ZERO, &pctx)) {
+			inode.i_dir_acl = 0;
+			inode_modified++;
+		} else
+			not_fixed++;
 	}
+
 	if (inode_modified)
 		e2fsck_write_inode(ctx, ino, &inode, "process_bad_inode");
+	if (!not_fixed)
+		ext2fs_unmark_inode_bitmap(ctx->inode_bad_map, ino);
 	return 0;
 }
 
@@ -1287,7 +1316,8 @@ extern int e2fsck_process_bad_inode(e2fsck_t ctx, ext2_ino_t dir,
  */
 static int allocate_dir_block(e2fsck_t ctx,
 			      struct ext2_db_entry *db,
-			      char *buf, struct problem_context *pctx)
+			      char *buf EXT2FS_ATTR((unused)), 
+			      struct problem_context *pctx)
 {
 	ext2_filsys fs = ctx->fs;
 	blk_t			blk;
@@ -1332,7 +1362,7 @@ static int allocate_dir_block(e2fsck_t ctx,
 	}
 
 	pctx->errcode = ext2fs_write_dir_block(fs, blk, block);
-	ext2fs_free_mem((void **) &block);
+	ext2fs_free_mem(&block);
 	if (pctx->errcode) {
 		pctx->str = "ext2fs_write_dir_block";
 		fix_problem(ctx, PR_2_ALLOC_DIRBOCK, pctx);
@@ -1366,11 +1396,11 @@ static int allocate_dir_block(e2fsck_t ctx,
 /*
  * This is a helper function for allocate_dir_block().
  */
-static int update_dir_block(ext2_filsys fs,
+static int update_dir_block(ext2_filsys fs EXT2FS_ATTR((unused)),
 			    blk_t	*block_nr,
 			    e2_blkcnt_t blockcnt,
-			    blk_t ref_block,
-			    int ref_offset, 
+			    blk_t ref_block EXT2FS_ATTR((unused)),
+			    int ref_offset EXT2FS_ATTR((unused)), 
 			    void *priv_data)
 {
 	struct ext2_db_entry *db;
