@@ -39,6 +39,12 @@ extern int optind;
 #ifdef HAVE_MALLOC_H
 #include <malloc.h>
 #endif
+#ifdef HAVE_SYS_TYPES_H
+#include <sys/types.h>
+#endif
+#ifdef HAVE_DIRENT_H
+#include <dirent.h>
+#endif
 
 #include "et/com_err.h"
 #include "e2fsck.h"
@@ -46,14 +52,15 @@ extern int optind;
 #include "../version.h"
 
 /* Command line options */
-static int swapfs = 0;
-static int normalize_swapfs = 0;
-static int cflag = 0;		/* check disk */
-static int show_version_only = 0;
-static int verbose = 0;
+static int swapfs;
+static int normalize_swapfs;
+static int cflag;		/* check disk */
+static int show_version_only;
+static int verbose;
 
-static int replace_bad_blocks = 0;
-static char *bad_blocks_file = 0;
+static int replace_bad_blocks;
+static int keep_bad_blocks;
+static char *bad_blocks_file;
 
 e2fsck_t e2fsck_global_ctx;	/* Try your very best not to use this! */
 
@@ -186,7 +193,7 @@ static void check_mount(e2fsck_t ctx)
 	}
 
 	printf(_("%s is mounted.  "), ctx->filesystem_name);
-	if (!isatty(0) || !isatty(1))
+	if (!ctx->interactive)
 		fatal_error(ctx, _("Cannot continue, aborting.\n\n"));
 	printf(_("\n\n\007\007\007\007WARNING!!!  "
 	       "Running e2fsck on a mounted filesystem may cause\n"
@@ -199,6 +206,44 @@ static void check_mount(e2fsck_t ctx)
 	return;
 }
 
+static int is_on_batt(void)
+{
+	FILE	*f;
+	DIR	*d;
+	char	tmp[80], tmp2[80], fname[80];
+	unsigned int	acflag;
+	struct dirent*	de;
+
+	f = fopen("/proc/apm", "r");
+	if (f) {
+		if (fscanf(f, "%s %s %s %x", tmp, tmp, tmp, &acflag) != 4) 
+			acflag = 1;
+		fclose(f);
+		return (acflag != 1);
+	}
+	d = opendir("/proc/acpi/ac_adapter");
+	if (d) {
+		while ((de=readdir(d)) != NULL) {
+			if (!strncmp(".", de->d_name, 1))
+				continue;
+			snprintf(fname, 80, "/proc/acpi/ac_adapter/%s/state", 
+				 de->d_name);
+			f = fopen(fname, "r");
+			if (!f)
+				continue;
+			if (fscanf(f, "%s %s", tmp2, tmp) != 2)
+				tmp[0] = 0;
+			fclose(f);
+			if (strncmp(tmp, "off-line", 8) == 0) {
+				closedir(d);
+				return 1;
+			}
+		}
+		closedir(d);
+	}
+	return 0;
+}
+
 /*
  * This routine checks to see if a filesystem can be skipped; if so,
  * it will exit with E2FSCK_OK.  Under some conditions it will print a
@@ -209,12 +254,16 @@ static void check_if_skip(e2fsck_t ctx)
 	ext2_filsys fs = ctx->fs;
 	const char *reason = NULL;
 	unsigned int reason_arg = 0;
+	long next_check;
+	int batt = is_on_batt();
+	time_t now = time(0);
 	
 	if ((ctx->options & E2F_OPT_FORCE) || bad_blocks_file ||
 	    cflag || swapfs)
 		return;
 	
-	if (fs->super->s_state & EXT2_ERROR_FS)
+	if ((fs->super->s_state & EXT2_ERROR_FS) ||
+	    !ext2fs_test_valid(fs))
 		reason = _(" contains a file system with errors");
 	else if ((fs->super->s_state & EXT2_VALID_FS) == 0)
 		reason = _(" was not cleanly unmounted");
@@ -223,11 +272,17 @@ static void check_if_skip(e2fsck_t ctx)
 		  (unsigned) fs->super->s_max_mnt_count)) {
 		reason = _(" has been mounted %u times without being checked");
 		reason_arg = fs->super->s_mnt_count;
+		if (batt && (fs->super->s_mnt_count < 
+			     (unsigned) fs->super->s_max_mnt_count*2))
+			reason = 0;
 	} else if (fs->super->s_checkinterval &&
-		 time(0) >= (fs->super->s_lastcheck +
-			     fs->super->s_checkinterval)) {
+		   ((now - fs->super->s_lastcheck) >= 
+		    fs->super->s_checkinterval)) {
 		reason = _(" has gone %u days without being checked");
-		reason_arg = (time(0) - fs->super->s_lastcheck)/(3600*24);
+		reason_arg = (now - fs->super->s_lastcheck)/(3600*24);
+		if (batt && ((now - fs->super->s_lastcheck) < 
+			     fs->super->s_checkinterval*2))
+			reason = 0;
 	}
 	if (reason) {
 		fputs(ctx->device_name, stdout);
@@ -235,11 +290,27 @@ static void check_if_skip(e2fsck_t ctx)
 		fputs(_(", check forced.\n"), stdout);
 		return;
 	}
-	printf(_("%s: clean, %d/%d files, %d/%d blocks\n"), ctx->device_name,
+	printf(_("%s: clean, %d/%d files, %d/%d blocks"), ctx->device_name,
 	       fs->super->s_inodes_count - fs->super->s_free_inodes_count,
 	       fs->super->s_inodes_count,
 	       fs->super->s_blocks_count - fs->super->s_free_blocks_count,
 	       fs->super->s_blocks_count);
+	next_check = 100000;
+	if (fs->super->s_max_mnt_count > 0) {
+		next_check = fs->super->s_max_mnt_count - fs->super->s_mnt_count;
+		if (next_check <= 0) 
+			next_check = 1;
+	}
+	if (fs->super->s_checkinterval &&
+	    ((now - fs->super->s_lastcheck) >= fs->super->s_checkinterval))
+		next_check = 1;
+	if (next_check <= 5) {
+		if (next_check == 1)
+			fputs(_(" (check after next mount)"), stdout);
+		else
+			printf(_(" (check in %ld mounts)"), next_check);
+	}
+	fputc('\n', stdout);
 	ext2fs_close(fs);
 	ctx->fs = NULL;
 	e2fsck_free_context(ctx);
@@ -256,12 +327,7 @@ struct percent_tbl {
 struct percent_tbl e2fsck_tbl = {
 	5, { 0, 70, 90, 92,  95, 100 }
 };
-static char bar[] =
-	"==============================================================="
-	"===============================================================";
-static char spaces[] =
-	"                                                               "
-	"                                                               ";
+static char bar[128], spaces[128];
 
 static float calc_percent(struct percent_tbl *tbl, int pass, int curr,
 			  int max)
@@ -282,7 +348,8 @@ extern void e2fsck_clear_progbar(e2fsck_t ctx)
 	if (!(ctx->flags & E2F_FLAG_PROG_BAR))
 		return;
 	
-	printf("\001%s\r\002", spaces + (sizeof(spaces) - 80));
+	printf("%s%s\r%s", ctx->start_meta, spaces + (sizeof(spaces) - 80),
+	       ctx->stop_meta);
 	fflush(stdout);
 	ctx->flags &= ~E2F_FLAG_PROG_BAR;
 }
@@ -292,9 +359,10 @@ int e2fsck_simple_progress(e2fsck_t ctx, const char *label, float percent,
 {
 	static const char spinner[] = "\\|/-";
 	int	i;
-	int	tick;
+	unsigned int	tick;
 	struct timeval	tv;
 	int dpywidth;
+	int fixed_percent;
 
 	if (ctx->flags & E2F_FLAG_PROG_SUPPRESS)
 		return 0;
@@ -304,9 +372,10 @@ int e2fsck_simple_progress(e2fsck_t ctx, const char *label, float percent,
 	 * percentage hasn't changed, then we skip out right
 	 * away. 
 	 */
-	if (ctx->progress_last_percent == (int) 10 * percent)
+	fixed_percent = (int) ((10 * percent) + 0.5);
+	if (ctx->progress_last_percent == fixed_percent)
 		return 0;
-	ctx->progress_last_percent = (int) 10 * percent;
+	ctx->progress_last_percent = fixed_percent;
 
 	/*
 	 * If we've already updated the spinner once within
@@ -316,7 +385,7 @@ int e2fsck_simple_progress(e2fsck_t ctx, const char *label, float percent,
 	gettimeofday(&tv, NULL);
 	tick = (tv.tv_sec << 3) + (tv.tv_usec / (1000000 / 8));
 	if ((tick == ctx->progress_last_time) &&
-	    (percent != 0.0) && (percent != 100.0))
+	    (fixed_percent != 0) && (fixed_percent != 1000))
 		return 0;
 	ctx->progress_last_time = tick;
 
@@ -333,18 +402,21 @@ int e2fsck_simple_progress(e2fsck_t ctx, const char *label, float percent,
 		dpywidth -= 8;
 
 	i = ((percent * dpywidth) + 50) / 100;
-	printf("\001%s: |%s%s", label, bar + (sizeof(bar) - (i+1)),
+	printf("%s%s: |%s%s", ctx->start_meta, label,
+	       bar + (sizeof(bar) - (i+1)),
 	       spaces + (sizeof(spaces) - (dpywidth - i + 1)));
-	if (percent == 100.0)
+	if (fixed_percent == 1000)
 		fputc('|', stdout);
 	else
 		fputc(spinner[ctx->progress_pos & 3], stdout);
+	printf(" %4.1f%%  ", percent);
 	if (dpynum)
-		printf(" %4.1f%%  %u\r\002", percent, dpynum);
+		printf("%u\r", dpynum);
 	else
-		printf(" %4.1f%%   \r\002", percent);
+		fputs(" \r", stdout);
+	fputs(ctx->stop_meta, stdout);
 	
-	if (percent == 100.0)
+	if (fixed_percent == 1000)
 		e2fsck_clear_progbar(ctx);
 	fflush(stdout);
 
@@ -392,7 +464,7 @@ static void reserve_stdio_fds(void)
 }
 
 #ifdef HAVE_SIGNAL_H
-static void signal_progress_on(int sig)
+static void signal_progress_on(int sig EXT2FS_ATTR((unused)))
 {
 	e2fsck_t ctx = e2fsck_global_ctx;
 
@@ -403,7 +475,7 @@ static void signal_progress_on(int sig)
 	ctx->progress_fd = 0;
 }
 
-static void signal_progress_off(int sig)
+static void signal_progress_off(int sig EXT2FS_ATTR((unused)))
 {
 	e2fsck_t ctx = e2fsck_global_ctx;
 
@@ -414,7 +486,7 @@ static void signal_progress_off(int sig)
 	ctx->progress = 0;
 }
 
-static void signal_cancel(int sig)
+static void signal_cancel(int sig EXT2FS_ATTR((unused)))
 {
 	e2fsck_t ctx = e2fsck_global_ctx;
 
@@ -494,6 +566,14 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 
 	setvbuf(stdout, NULL, _IONBF, BUFSIZ);
 	setvbuf(stderr, NULL, _IONBF, BUFSIZ);
+	if (isatty(0) && isatty(1)) {
+		ctx->interactive = 1;
+	} else {
+		ctx->start_meta[0] = '\001';
+		ctx->stop_meta[0] = '\002';
+	}
+	memset(bar, '=', sizeof(bar)-1);
+	memset(spaces, ' ', sizeof(spaces)-1);
 	initialize_ext2_error_table();
 	blkid_get_cache(&ctx->blkid, NULL);
 	
@@ -501,7 +581,7 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 		ctx->program_name = *argv;
 	else
 		ctx->program_name = "e2fsck";
-	while ((c = getopt (argc, argv, "panyrcC:B:dE:fvtFVM:b:I:j:P:l:L:N:SsD")) != EOF)
+	while ((c = getopt (argc, argv, "panyrcC:B:dE:fvtFVM:b:I:j:P:l:L:N:SsDk")) != EOF)
 		switch (c) {
 		case 'C':
 			ctx->progress = e2fsck_update_progress;
@@ -622,6 +702,9 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 					  "of e2fsck\n"));
 			exit(1);
 #endif
+		case 'k':
+			keep_bad_blocks++;
+			break;
 		default:
 			usage(ctx);
 		}
@@ -632,7 +715,15 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 	if ((ctx->options & E2F_OPT_NO) && !bad_blocks_file &&
 	    !cflag && !swapfs && !(ctx->options & E2F_OPT_COMPRESS_DIRS))
 		ctx->options |= E2F_OPT_READONLY;
+	ctx->io_options = strchr(argv[optind], '?');
+	if (ctx->io_options) 
+		*ctx->io_options++ = 0;
 	ctx->filesystem_name = blkid_get_devname(ctx->blkid, argv[optind], 0);
+	if (!ctx->filesystem_name) {
+		com_err(ctx->program_name, 0, _("Unable to resolve '%s'"), 
+			argv[optind]);
+		fatal_error(ctx, 0);
+	}
 	if (extended_opts)
 		parse_extended_opts(ctx, extended_opts);
 	
@@ -772,7 +863,7 @@ int main (int argc, char *argv[])
 	if (!(ctx->options & E2F_OPT_PREEN) &&
 	    !(ctx->options & E2F_OPT_NO) &&
 	    !(ctx->options & E2F_OPT_YES)) {
-		if (!isatty (0) || !isatty (1))
+		if (!ctx->interactive)
 			fatal_error(ctx,
 				    _("need terminal for interactive repairs"));
 	}
@@ -789,22 +880,23 @@ restart:
 		flags |= EXT2_FLAG_RW;
 
 	if (ctx->superblock && ctx->blocksize) {
-		retval = ext2fs_open(ctx->filesystem_name, flags,
-				     ctx->superblock, ctx->blocksize,
-				     io_ptr, &fs);
+		retval = ext2fs_open2(ctx->filesystem_name, ctx->io_options, 
+				      flags, ctx->superblock, ctx->blocksize,
+				      io_ptr, &fs);
 	} else if (ctx->superblock) {
 		int blocksize;
 		for (blocksize = EXT2_MIN_BLOCK_SIZE;
 		     blocksize <= EXT2_MAX_BLOCK_SIZE; blocksize *= 2) {
-			retval = ext2fs_open(ctx->filesystem_name, flags,
-					     ctx->superblock, blocksize,
-					     io_ptr, &fs);
+			retval = ext2fs_open2(ctx->filesystem_name, 
+					      ctx->io_options, flags,
+					      ctx->superblock, blocksize,
+					      io_ptr, &fs);
 			if (!retval)
 				break;
 		}
 	} else 
-		retval = ext2fs_open(ctx->filesystem_name, flags, 
-				     0, 0, io_ptr, &fs);
+		retval = ext2fs_open2(ctx->filesystem_name, ctx->io_options, 
+				      flags, 0, 0, io_ptr, &fs);
 	if (!ctx->superblock && !(ctx->options & E2F_OPT_PREEN) &&
 	    !(ctx->flags & E2F_FLAG_SB_SPECIFIED) &&
 	    ((retval == EXT2_ET_BAD_MAGIC) ||
@@ -969,6 +1061,7 @@ restart:
 
 	if (ctx->superblock)
 		set_latch_flags(PR_LATCH_RELOC, PRL_LATCHED, 0);
+	ext2fs_mark_valid(fs);
 	check_super_block(ctx);
 	if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
 		fatal_error(ctx, 0);
@@ -976,7 +1069,7 @@ restart:
 	if (bad_blocks_file)
 		read_bad_blocks_file(ctx, bad_blocks_file, replace_bad_blocks);
 	else if (cflag)
-		test_disk(ctx);
+		read_bad_blocks_file(ctx, 0, !keep_bad_blocks); /* Test disk */
 	if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
 		fatal_error(ctx, 0);
 #ifdef ENABLE_SWAPFS
