@@ -149,8 +149,7 @@ static void print_mntopts(struct ext2_super_block * s, FILE *f)
 void list_super2(struct ext2_super_block * sb, FILE *f)
 {
 	int inode_blocks_per_group;
-	char buf[80];
-	const char *os;
+	char buf[80], *str;
 	time_t	tm;
 
 	inode_blocks_per_group = (((sb->s_inodes_per_group *
@@ -188,13 +187,9 @@ void list_super2(struct ext2_super_block * sb, FILE *f)
 	fprintf(f, "Errors behavior:          ");
 	print_fs_errors(f, sb->s_errors);
 	fprintf(f, "\n");
-	switch (sb->s_creator_os) {
-	    case EXT2_OS_LINUX: os = "Linux"; break;
-	    case EXT2_OS_HURD:  os = "GNU/Hurd"; break;
-	    case EXT2_OS_MASIX: os = "Masix"; break;
-	    default:		os = "unknown"; break;
-	}
-	fprintf(f, "Filesystem OS type:       %s\n", os);
+	str = e2p_os2string(sb->s_creator_os);
+	fprintf(f, "Filesystem OS type:       %s\n", str);
+	free(str);
 	fprintf(f, "Inode count:              %u\n", sb->s_inodes_count);
 	fprintf(f, "Block count:              %u\n", sb->s_blocks_count);
 	fprintf(f, "Reserved block count:     %u\n", sb->s_r_blocks_count);
@@ -203,6 +198,9 @@ void list_super2(struct ext2_super_block * sb, FILE *f)
 	fprintf(f, "First block:              %u\n", sb->s_first_data_block);
 	fprintf(f, "Block size:               %u\n", EXT2_BLOCK_SIZE(sb));
 	fprintf(f, "Fragment size:            %u\n", EXT2_FRAG_SIZE(sb));
+	if (sb->s_reserved_gdt_blocks)
+		fprintf(f, "Reserved GDT blocks:      %u\n", 
+			sb->s_reserved_gdt_blocks);
 	fprintf(f, "Blocks per group:         %u\n", sb->s_blocks_per_group);
 	fprintf(f, "Fragments per group:      %u\n", sb->s_frags_per_group);
 	fprintf(f, "Inodes per group:         %u\n", sb->s_inodes_per_group);
@@ -259,6 +257,16 @@ void list_super2(struct ext2_super_block * sb, FILE *f)
 	if (!e2p_is_null_uuid(sb->s_hash_seed))
 		fprintf(f, "Directory Hash Seed:      %s\n",
 			e2p_uuid2str(sb->s_hash_seed));
+	if (sb->s_jnl_backup_type) {
+		fprintf(f, "Journal backup:           ");
+		switch (sb->s_jnl_backup_type) {
+		case 1:
+			fprintf(f, "inode blocks\n");
+			break;
+		default:
+			fprintf(f, "type %u\n", sb->s_jnl_backup_type);
+		}
+	}
 }
 
 void list_super (struct ext2_super_block * s)
