@@ -37,7 +37,7 @@ int ext2fs_set_bit(int nr,void * addr)
 
 	ADDR += nr >> 3;
 	mask = 1 << (nr & 0x07);
-	retval = (mask & *ADDR) != 0;
+	retval = mask & *ADDR;
 	*ADDR |= mask;
 	return retval;
 }
@@ -49,7 +49,7 @@ int ext2fs_clear_bit(int nr, void * addr)
 
 	ADDR += nr >> 3;
 	mask = 1 << (nr & 0x07);
-	retval = (mask & *ADDR) != 0;
+	retval = mask & *ADDR;
 	*ADDR &= ~mask;
 	return retval;
 }
@@ -61,7 +61,7 @@ int ext2fs_test_bit(int nr, const void * addr)
 
 	ADDR += nr >> 3;
 	mask = 1 << (nr & 0x07);
-	return ((mask & *ADDR) != 0);
+	return (mask & *ADDR);
 }
 
 #endif	/* !_EXT2_HAVE_ASM_BITOPS_ */
@@ -71,9 +71,9 @@ void ext2fs_warn_bitmap(errcode_t errcode, unsigned long arg,
 {
 #ifndef OMIT_COM_ERR
 	if (description)
-		com_err(0, errcode, "#%u for %s", arg, description);
+		com_err(0, errcode, "#%lu for %s", arg, description);
 	else
-		com_err(0, errcode, "#%u", arg);
+		com_err(0, errcode, "#%lu", arg);
 #endif
 }
 
@@ -83,9 +83,9 @@ void ext2fs_warn_bitmap2(ext2fs_generic_bitmap bitmap,
 #ifndef OMIT_COM_ERR
 	if (bitmap->description)
 		com_err(0, bitmap->base_error_code+code,
-			"#%u for %s", arg, bitmap->description);
+			"#%lu for %s", arg, bitmap->description);
 	else
-		com_err(0, bitmap->base_error_code + code, "#%u", arg);
+		com_err(0, bitmap->base_error_code + code, "#%lu", arg);
 #endif
 }
 
