@@ -38,8 +38,10 @@ struct read_bb_record {
  #pragma argsused
 #endif
 static int mark_bad_block(ext2_filsys fs, blk_t *block_nr,
-			  e2_blkcnt_t blockcnt, blk_t ref_block,
-			  int ref_offset, void *priv_data)
+			  e2_blkcnt_t blockcnt EXT2FS_ATTR((unused)), 
+			  blk_t ref_block EXT2FS_ATTR((unused)),
+			  int ref_offset EXT2FS_ATTR((unused)), 
+			  void *priv_data)
 {
 	struct read_bb_record *rb = (struct read_bb_record *) priv_data;
 	
