@@ -39,6 +39,7 @@ struct test_private_data {
 	int flags;
 	FILE *outfile;
 	unsigned long block;
+	int read_abort_count, write_abort_count;
 	void (*read_blk)(unsigned long block, int count, errcode_t err);
 	void (*write_blk)(unsigned long block, int count, errcode_t err);
 	void (*set_blksize)(int blksize, errcode_t err);
@@ -55,6 +56,8 @@ static errcode_t test_write_blk(io_channel channel, unsigned long block,
 static errcode_t test_flush(io_channel channel);
 static errcode_t test_write_byte(io_channel channel, unsigned long offset,
 				 int count, const void *buf);
+static errcode_t test_set_option(io_channel channel, const char *option, 
+				 const char *arg);
 
 static struct struct_io_manager struct_test_manager = {
 	EXT2_ET_MAGIC_IO_MANAGER,
@@ -65,8 +68,8 @@ static struct struct_io_manager struct_test_manager = {
 	test_read_blk,
 	test_write_blk,
 	test_flush,
-	test_write_byte
-	
+	test_write_byte,
+	test_set_option
 };
 
 io_manager test_io_manager = &struct_test_manager;
@@ -92,6 +95,8 @@ void (*test_io_cb_write_byte)
 #define TEST_FLAG_WRITE			0x02
 #define TEST_FLAG_SET_BLKSIZE		0x04
 #define TEST_FLAG_FLUSH			0x08
+#define TEST_FLAG_DUMP			0x10
+#define TEST_FLAG_SET_OPTION		0x20
 
 static void test_dump_block(io_channel channel,
 			    struct test_private_data *data,
@@ -113,28 +118,41 @@ static void test_dump_block(io_channel channel,
 	}
 }
 
+static void test_abort(io_channel channel, unsigned long block)
+{
+	struct test_private_data *data;
+	FILE *f;
+
+	data = (struct test_private_data *) channel->private_data;
+	f = data->outfile;
+	test_flush(channel);
+
+	fprintf(f, "Aborting due to I/O to block %lu\n", block);
+	fflush(f);
+	abort();
+}
+
 static errcode_t test_open(const char *name, int flags, io_channel *channel)
 {
 	io_channel	io = NULL;
 	struct test_private_data *data = NULL;
 	errcode_t	retval;
+	char		*value;
 
 	if (name == 0)
 		return EXT2_ET_BAD_DEVICE_NAME;
-	retval = ext2fs_get_mem(sizeof(struct struct_io_channel),
-				(void **) &io);
+	retval = ext2fs_get_mem(sizeof(struct struct_io_channel), &io);
 	if (retval)
 		return retval;
 	memset(io, 0, sizeof(struct struct_io_channel));
 	io->magic = EXT2_ET_MAGIC_IO_CHANNEL;
-	retval = ext2fs_get_mem(sizeof(struct test_private_data),
-				(void **) &data);
+	retval = ext2fs_get_mem(sizeof(struct test_private_data), &data);
 	if (retval) {
 		retval = EXT2_ET_NO_MEMORY;
 		goto cleanup;
 	}
 	io->manager = test_io_manager;
-	retval = ext2fs_get_mem(strlen(name)+1, (void **) &io->name);
+	retval = ext2fs_get_mem(strlen(name)+1, &io->name);
 	if (retval)
 		goto cleanup;
 
@@ -160,27 +178,35 @@ static errcode_t test_open(const char *name, int flags, io_channel *channel)
 	data->write_byte = 	test_io_cb_write_byte;
 
 	data->outfile = NULL;
-	if (getenv("TEST_IO_LOGFILE"))
-		data->outfile = fopen(getenv("TEST_IO_LOGFILE"), "w");
+	if ((value = getenv("TEST_IO_LOGFILE")) != NULL)
+		data->outfile = fopen(value, "w");
 	if (!data->outfile)
 		data->outfile = stderr;
 
 	data->flags = 0;
-	if (getenv("TEST_IO_FLAGS"))
-		data->flags = strtoul(getenv("TEST_IO_FLAGS"), NULL, 0);
+	if ((value = getenv("TEST_IO_FLAGS")) != NULL)
+		data->flags = strtoul(value, NULL, 0);
 	
 	data->block = 0;
-	if (getenv("TEST_IO_BLOCK"))
-		data->block = strtoul(getenv("TEST_IO_BLOCK"), NULL, 0);
+	if ((value = getenv("TEST_IO_BLOCK")) != NULL)
+		data->block = strtoul(value, NULL, 0);
+
+	data->read_abort_count = 0;
+	if ((value = getenv("TEST_IO_READ_ABORT")) != NULL)
+		data->read_abort_count = strtoul(value, NULL, 0);
+
+	data->write_abort_count = 0;
+	if ((value = getenv("TEST_IO_WRITE_ABORT")) != NULL)
+		data->write_abort_count = strtoul(value, NULL, 0);
 	
 	*channel = io;
 	return 0;
 
 cleanup:
 	if (io)
-		ext2fs_free_mem((void **) &io);
+		ext2fs_free_mem(&io);
 	if (data)
-		ext2fs_free_mem((void **) &data);
+		ext2fs_free_mem(&data);
 	return retval;
 }
 
@@ -202,10 +228,10 @@ static errcode_t test_close(io_channel channel)
 	if (data->outfile && data->outfile != stderr)
 		fclose(data->outfile);
 	
-	ext2fs_free_mem((void **) &channel->private_data);
+	ext2fs_free_mem(&channel->private_data);
 	if (channel->name)
-		ext2fs_free_mem((void **) &channel->name);
-	ext2fs_free_mem((void **) &channel);
+		ext2fs_free_mem(&channel->name);
+	ext2fs_free_mem(&channel);
 	return retval;
 }
 
@@ -249,8 +275,12 @@ static errcode_t test_read_blk(io_channel channel, unsigned long block,
 		fprintf(data->outfile,
 			"Test_io: read_blk(%lu, %d) returned %s\n",
 			block, count, retval ? error_message(retval) : "OK");
-	if (data->block && data->block == block)
-		test_dump_block(channel, data, block, buf);
+	if (data->block && data->block == block) {
+		if (data->flags & TEST_FLAG_DUMP)
+			test_dump_block(channel, data, block, buf);
+		if (--data->read_abort_count == 0)
+			test_abort(channel, block);
+	} 
 	return retval;
 }
 
@@ -272,8 +302,12 @@ static errcode_t test_write_blk(io_channel channel, unsigned long block,
 		fprintf(data->outfile,
 			"Test_io: write_blk(%lu, %d) returned %s\n",
 			block, count, retval ? error_message(retval) : "OK");
-	if (data->block && data->block == block)
-		test_dump_block(channel, data, block, buf);
+	if (data->block && data->block == block) {
+		if (data->flags & TEST_FLAG_DUMP)
+			test_dump_block(channel, data, block, buf);
+		if (--data->write_abort_count == 0)
+			test_abort(channel, block);
+	}
 	return retval;
 }
 
@@ -320,3 +354,29 @@ static errcode_t test_flush(io_channel channel)
 	return retval;
 }
 
+static errcode_t test_set_option(io_channel channel, const char *option, 
+				 const char *arg)
+{
+	struct test_private_data *data;
+	errcode_t	retval = 0;
+
+	EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
+	data = (struct test_private_data *) channel->private_data;
+	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_TEST_IO_CHANNEL);
+
+
+	if (data->flags & TEST_FLAG_SET_OPTION)
+		fprintf(data->outfile, "Test_io: set_option(%s, %s) ", 
+			option, arg);
+	if (data->real && data->real->manager->set_option) {
+		retval = (data->real->manager->set_option)(data->real, 
+							   option, arg);
+		if (data->flags & TEST_FLAG_SET_OPTION)
+			fprintf(data->outfile, "returned %s\n",
+				retval ? error_message(retval) : "OK");
+	} else {
+		if (data->flags & TEST_FLAG_SET_OPTION)
+			fprintf(data->outfile, "not implemented\n");
+	}
+	return retval;
+}
