@@ -141,8 +141,8 @@ static int bb_output (unsigned long bad)
 
 static void print_status(void)
 {
-	fprintf(stderr, "%9ld/%9ld", currently_testing, num_blocks);
-	fprintf(stderr, "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
+	fprintf(stderr, "%15ld/%15ld", currently_testing, num_blocks);
+	fputs("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b", stderr);
 	fflush (stderr);
 }
 
@@ -152,9 +152,7 @@ static void alarm_intr(int alnum EXT2FS_ATTR((unused)))
 	alarm(1);
 	if (!num_blocks)
 		return;
-	fprintf(stderr, "%9ld/%9ld", currently_testing, num_blocks);
-	fprintf(stderr, "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
-	fflush (stderr);
+	print_status();
 }
 
 static void *terminate_addr = NULL;
@@ -412,7 +410,7 @@ static unsigned int test_ro (int dev, unsigned long last_block,
 	num_blocks = 0;
 	alarm(0);
 	if (s_flag || v_flag)
-		fputs(done_string, stderr);
+		fputs(_(done_string), stderr);
 
 	fflush (stderr);
 	free (blkbuf);
@@ -493,7 +491,7 @@ static unsigned int test_rw (int dev, unsigned long last_block,
 		num_blocks = 0;
 		alarm (0);
 		if (s_flag | v_flag)
-			fputs(done_string, stderr);
+			fputs(_(done_string), stderr);
 		flush_bufs();
 		if (s_flag | v_flag)
 			fputs(_("Reading and comparing: "), stderr);
@@ -534,7 +532,7 @@ static unsigned int test_rw (int dev, unsigned long last_block,
 		num_blocks = 0;
 		alarm (0);
 		if (s_flag | v_flag)
-			fputs(done_string, stderr);
+			fputs(_(done_string), stderr);
 		flush_bufs();
 	}
 	uncapture_terminate();
@@ -756,7 +754,7 @@ static unsigned int test_nd (int dev, unsigned long last_block,
 		num_blocks = 0;
 		alarm(0);
 		if (s_flag || v_flag > 1)
-			fputs(done_string, stderr);
+			fputs(_(done_string), stderr);
 
 		flush_bufs();
 	}
