@@ -47,7 +47,8 @@ char * device_name = NULL;
 
 static void usage(void)
 {
-	fprintf(stderr, _("Usage: %s [-r] device file\n"), program_name);
+	fprintf(stderr, _("Usage: %s [-rsI] device image_file\n"), 
+		program_name);
 	exit (1);
 }
 
@@ -58,7 +59,7 @@ static void write_header(int fd, struct ext2_image_hdr *hdr, int blocksize)
 
 	header_buf = malloc(blocksize);
 	if (!header_buf) {
-		fprintf(stderr, _("Couldn't allocate header buffer\n"));
+		fputs(_("Couldn't allocate header buffer\n"), stderr);
 		exit(1);
 	}
 
@@ -146,9 +147,11 @@ static void write_image_file(ext2_filsys fs, int fd)
  * These set of functions are used to write a RAW image file.
  */
 ext2fs_block_bitmap meta_block_map;
+ext2fs_block_bitmap scramble_block_map;	/* Directory blocks to be scrambled */
 
 struct process_block_struct {
 	ext2_ino_t	ino;
+	int		is_dir;
 };
 
 /*
@@ -160,8 +163,9 @@ struct process_block_struct {
 static ino_t stashed_ino = 0;
 static struct ext2_inode *stashed_inode;
 
-static errcode_t meta_get_blocks(ext2_filsys fs, ext2_ino_t ino,
-				  blk_t *blocks)
+static errcode_t meta_get_blocks(ext2_filsys fs EXT2FS_ATTR((unused)), 
+				 ext2_ino_t ino,
+				 blk_t *blocks)
 {
 	int	i;
 	
@@ -173,7 +177,8 @@ static errcode_t meta_get_blocks(ext2_filsys fs, ext2_ino_t ino,
 	return 0;
 }
 
-static errcode_t meta_check_directory(ext2_filsys fs, ext2_ino_t ino)
+static errcode_t meta_check_directory(ext2_filsys fs EXT2FS_ATTR((unused)), 
+				      ext2_ino_t ino)
 {
 	if ((ino != stashed_ino) || !stashed_inode)
 		return EXT2_ET_CALLBACK_NOTHANDLED;
@@ -183,7 +188,8 @@ static errcode_t meta_check_directory(ext2_filsys fs, ext2_ino_t ino)
 	return 0;
 }
 
-static errcode_t meta_read_inode(ext2_filsys fs, ext2_ino_t ino,
+static errcode_t meta_read_inode(ext2_filsys fs EXT2FS_ATTR((unused)), 
+				 ext2_ino_t ino,
 				 struct ext2_inode *inode)
 {
 	if ((ino != stashed_ino) || !stashed_inode)
@@ -206,17 +212,29 @@ static void use_inode_shortcuts(ext2_filsys fs, int bool)
 	}
 }
 
-static int process_dir_block(ext2_filsys fs, blk_t *block_nr,
-			     e2_blkcnt_t blockcnt, blk_t ref_block,
-			     int ref_offset, void *priv_data)
+static int process_dir_block(ext2_filsys fs EXT2FS_ATTR((unused)), 
+			     blk_t *block_nr,
+			     e2_blkcnt_t blockcnt EXT2FS_ATTR((unused)), 
+			     blk_t ref_block EXT2FS_ATTR((unused)),
+			     int ref_offset EXT2FS_ATTR((unused)), 
+			     void *priv_data EXT2FS_ATTR((unused)))
 {
+	struct process_block_struct *p;
+
+	p = (struct process_block_struct *) priv_data;
+
 	ext2fs_mark_block_bitmap(meta_block_map, *block_nr);
+	if (scramble_block_map && p->is_dir && blockcnt >= 0) 
+		ext2fs_mark_block_bitmap(scramble_block_map, *block_nr);
 	return 0;
 }
 
-static int process_file_block(ext2_filsys fs, blk_t *block_nr,
-			     e2_blkcnt_t blockcnt, blk_t ref_block,
-			     int ref_offset, void *priv_data)
+static int process_file_block(ext2_filsys fs EXT2FS_ATTR((unused)), 
+			      blk_t *block_nr,
+			      e2_blkcnt_t blockcnt, 
+			      blk_t ref_block EXT2FS_ATTR((unused)),
+			      int ref_offset EXT2FS_ATTR((unused)), 
+			      void *priv_data EXT2FS_ATTR((unused)))
 {
 	if (blockcnt < 0) {
 		ext2fs_mark_block_bitmap(meta_block_map, *block_nr);
@@ -227,7 +245,7 @@ static int process_file_block(ext2_filsys fs, blk_t *block_nr,
 static void mark_table_blocks(ext2_filsys fs)
 {
 	blk_t	block, b;
-	int	i,j;
+	unsigned int	i,j;
 	
 	block = fs->super->s_first_data_block;
 	/*
@@ -249,7 +267,7 @@ static void mark_table_blocks(ext2_filsys fs)
 		 */
 		if (fs->group_desc[i].bg_inode_table) {
 			for (j = 0, b = fs->group_desc[i].bg_inode_table;
-			     j < fs->inode_blocks_per_group;
+			     j < (unsigned) fs->inode_blocks_per_group;
 			     j++, b++)
 				ext2fs_mark_block_bitmap(meta_block_map, b);
 		}
@@ -317,14 +335,81 @@ static void write_block(int fd, char *buf, int sparse_offset,
 	}
 }
 
+int name_id[256];
+
+static void scramble_dir_block(ext2_filsys fs, blk_t blk, char *buf)
+{
+	char *p, *end, *cp;
+	struct ext2_dir_entry_2 *dirent;
+	int rec_len, id, len;
+
+	end = buf + fs->blocksize;
+	for (p = buf; p < end-8; p += rec_len) {
+		dirent = (struct ext2_dir_entry_2 *) p;
+		rec_len = dirent->rec_len;
+#ifdef EXT2FS_ENABLE_SWAPFS
+		if (fs->flags & EXT2_FLAG_SWAP_BYTES) 
+			rec_len = ext2fs_swab16(rec_len);
+#endif
+#if 0
+		printf("rec_len = %d, name_len = %d\n", rec_len, dirent->name_len);
+#endif
+		if (rec_len < 8 || (rec_len % 4) ||
+		    (p+rec_len > end)) {
+			printf("Corrupt directory block %lu: "
+			       "bad rec_len (%d)\n", blk, rec_len);
+			rec_len = end - p;
+#ifdef EXT2FS_ENABLE_SWAPFS
+			if (fs->flags & EXT2_FLAG_SWAP_BYTES) 
+				dirent->rec_len = ext2fs_swab16(rec_len);
+#endif
+			continue;
+		}
+		if (dirent->name_len + 8 > rec_len) {
+			printf("Corrupt directory block %lu: "
+			       "bad name_len (%d)\n", blk, dirent->name_len);
+			dirent->name_len = rec_len - 8;
+			continue;
+		}
+		cp = p+8;
+		len = rec_len - dirent->name_len - 8;
+		if (len > 0)
+			memset(cp+dirent->name_len, 0, len);
+		if (dirent->name_len==1 && cp[0] == '.')
+			continue;
+		if (dirent->name_len==2 && cp[0] == '.' && cp[1] == '.')
+			continue;
+
+		memset(cp, 'A', dirent->name_len);
+		len = dirent->name_len;
+		id = name_id[len]++;
+		while ((len > 0) && (id > 0)) {
+			*cp += id % 26;
+			id = id / 26;
+			cp++;
+			len--;
+		}
+	}
+}
+
 static void output_meta_data_blocks(ext2_filsys fs, int fd)
 {
 	errcode_t	retval;
 	blk_t		blk;
-	char		buf[8192], zero_buf[8192];
+	char		*buf, *zero_buf;
 	int		sparse = 0;
 
-	memset(zero_buf, 0, sizeof(zero_buf));
+	buf = malloc(fs->blocksize);
+	if (!buf) {
+		com_err(program_name, ENOMEM, "while allocating buffer");
+		exit(1);
+	}
+	zero_buf = malloc(fs->blocksize);
+	if (!zero_buf) {
+		com_err(program_name, ENOMEM, "while allocating buffer");
+		exit(1);
+	}
+	memset(zero_buf, 0, fs->blocksize);
 	for (blk = 0; blk < fs->super->s_blocks_count; blk++) {
 		if ((blk >= fs->super->s_first_data_block) &&
 		    ext2fs_test_block_bitmap(meta_block_map, blk)) {
@@ -333,6 +418,9 @@ static void output_meta_data_blocks(ext2_filsys fs, int fd)
 				com_err(program_name, retval,
 					"error reading block %d", blk);
 			}
+			if (scramble_block_map && 
+			    ext2fs_test_block_bitmap(scramble_block_map, blk))
+				scramble_dir_block(fs, blk, buf);
 			if ((fd != 1) && check_zero_block(buf, fs->blocksize))
 				goto sparse_write;
 			write_block(fd, buf, sparse, fs->blocksize, blk);
@@ -354,7 +442,7 @@ static void output_meta_data_blocks(ext2_filsys fs, int fd)
 	write_block(fd, zero_buf, sparse, 1, -1);
 }
 
-static void write_raw_image_file(ext2_filsys fs, int fd)
+static void write_raw_image_file(ext2_filsys fs, int fd, int scramble_flag)
 {
 	struct process_block_struct	pb;
 	struct ext2_inode		inode;
@@ -369,6 +457,16 @@ static void write_raw_image_file(ext2_filsys fs, int fd)
 		com_err(program_name, retval, "while allocating block bitmap");
 		exit(1);
 	}
+
+	if (scramble_flag) {
+		retval = ext2fs_allocate_block_bitmap(fs, "scramble block map",
+						      &scramble_block_map);
+		if (retval) {
+			com_err(program_name, retval, 
+				"while allocating scramble block bitmap");
+			exit(1);
+		}
+	}
 	
 	mark_table_blocks(fs);
 
@@ -407,6 +505,8 @@ static void write_raw_image_file(ext2_filsys fs, int fd)
 			continue;
 		
 		stashed_ino = ino;
+		pb.ino = ino;
+		pb.is_dir = LINUX_S_ISDIR(inode.i_mode);
 		if (LINUX_S_ISDIR(inode.i_mode) ||
 		    (LINUX_S_ISLNK(inode.i_mode) &&
 		     ext2fs_inode_has_valid_blocks(&inode)) ||
@@ -438,14 +538,83 @@ static void write_raw_image_file(ext2_filsys fs, int fd)
 	output_meta_data_blocks(fs, fd);
 }
 
+static void install_image(char *device, char *image_fn, int raw_flag)
+{
+	errcode_t retval;
+	ext2_filsys fs;
+	int open_flag = EXT2_FLAG_IMAGE_FILE;
+	int fd = 0;
+	io_manager	io_ptr;
+	io_channel	io, image_io;
+
+	if (raw_flag) {
+		com_err(program_name, 0, "Raw images cannot be installed");
+		exit(1);
+	}
+	
+#ifdef CONFIG_TESTIO_DEBUG
+	io_ptr = test_io_manager;
+	test_io_backing_manager = unix_io_manager;
+#else
+	io_ptr = unix_io_manager;
+#endif
+
+	retval = ext2fs_open (image_fn, open_flag, 0, 0,
+			      io_ptr, &fs);
+        if (retval) {
+		com_err (program_name, retval, _("while trying to open %s"),
+			 image_fn);
+		exit(1);
+	}
+
+	retval = ext2fs_read_bitmaps (fs);
+	if (retval) {
+		com_err(program_name, retval, "error reading bitmaps");
+		exit(1);
+	}
+
+
+	fd = open(image_fn, O_RDONLY);
+	if (fd < 0) {
+		perror(image_fn);
+		exit(1);
+	}
+
+	retval = io_ptr->open(device, IO_FLAG_RW, &io); 
+	if (retval) {
+		com_err(device, 0, "while opening device file");
+		exit(1);
+	}
+
+	image_io = fs->io;
+
+	ext2fs_rewrite_to_io(fs, io);
+
+	if (lseek(fd, fs->image_header->offset_inode, SEEK_SET) < 0) {
+		perror("lseek");
+		exit(1);
+	}
+
+	retval = ext2fs_image_inode_read(fs, fd, 0);
+	if (retval) {
+		com_err(image_fn, 0, "while restoring the image table");
+		exit(1);
+	}
+
+	ext2fs_close (fs);
+	exit (0);
+}
+
 int main (int argc, char ** argv)
 {
 	int c;
 	errcode_t retval;
 	ext2_filsys fs;
-	char *outfn;
+	char *image_fn;
 	int open_flag = 0;
 	int raw_flag = 0;
+	int install_flag = 0;
+	int scramble_flag = 0;
 	int fd = 0;
 
 #ifdef ENABLE_NLS
@@ -459,34 +628,46 @@ int main (int argc, char ** argv)
 	if (argc && *argv)
 		program_name = *argv;
 	initialize_ext2_error_table();
-	while ((c = getopt (argc, argv, "r")) != EOF)
+	while ((c = getopt (argc, argv, "rsI")) != EOF)
 		switch (c) {
 		case 'r':
 			raw_flag++;
 			break;
+		case 's':
+			scramble_flag++;
+			break;
+		case 'I':
+			install_flag++;
+			break;
 		default:
 			usage();
 		}
 	if (optind != argc - 2 )
 		usage();
 	device_name = argv[optind];
-	outfn = argv[optind+1];
+	image_fn = argv[optind+1];
+
+	if (install_flag) {
+		install_image(device_name, image_fn, raw_flag);
+		exit (0);
+	}
+
 	retval = ext2fs_open (device_name, open_flag, 0, 0,
 			      unix_io_manager, &fs);
         if (retval) {
 		com_err (program_name, retval, _("while trying to open %s"),
 			 device_name);
-		printf(_("Couldn't find valid filesystem superblock.\n"));
+		fputs(_("Couldn't find valid filesystem superblock.\n"), stdout);
 		exit(1);
 	}
 
-	if (strcmp(outfn, "-") == 0)
+	if (strcmp(image_fn, "-") == 0)
 		fd = 1;
 	else {
 #ifdef HAVE_OPEN64
-		fd = open64(outfn, O_CREAT|O_TRUNC|O_WRONLY, 0600);
+		fd = open64(image_fn, O_CREAT|O_TRUNC|O_WRONLY, 0600);
 #else
-		fd = open(outfn, O_CREAT|O_TRUNC|O_WRONLY, 0600);
+		fd = open(image_fn, O_CREAT|O_TRUNC|O_WRONLY, 0600);
 #endif
 		if (fd < 0) {
 			com_err(program_name, errno,
@@ -496,7 +677,7 @@ int main (int argc, char ** argv)
 	}
 
 	if (raw_flag)
-		write_raw_image_file(fs, fd);
+		write_raw_image_file(fs, fd, scramble_flag);
 	else
 		write_image_file(fs, fd);
 
