@@ -1,7 +1,8 @@
 /*
  * mke2fs.c - Make a ext2fs filesystem.
  * 
- * Copyright (C) 1994, 1995, 1996, 1997 Theodore Ts'o.
+ * Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002,
+ * 	2003, 2004, 2005 by Theodore Ts'o.
  *
  * %Begin-Header%
  * This file may be redistributed under the terms of the GNU Public
@@ -85,6 +86,7 @@ char *journal_device;
 int sync_kludge;	/* Set using the MKE2FS_SYNC env. option */
 
 int sys_page_size = 4096;
+int linux_version_code = 0;
 
 static void usage(void)
 {
@@ -93,7 +95,7 @@ static void usage(void)
 	" [-N number-of-inodes]\n\t[-m reserved-blocks-percentage] "
 	"[-o creator-os] [-g blocks-per-group]\n\t[-L volume-label] "
 	"[-M last-mounted-directory] [-O feature[,...]]\n\t"
-	"[-r fs-revision] [-R raid_opts] [-qvSV] device [blocks-count]\n"),
+	"[-r fs-revision] [-R options] [-qvSV] device [blocks-count]\n"),
 		program_name);
 	exit(1);
 }
@@ -119,6 +121,30 @@ static int int_log10(unsigned int arg)
 	return l;
 }
 
+static int parse_version_number(const char *s)
+{
+	int	major, minor, rev;
+	char	*endptr;
+	const char *cp = s;
+
+	if (!s)
+		return 0;
+	major = strtol(cp, &endptr, 10);
+	if (cp == endptr || *endptr != '.')
+		return 0;
+	cp = endptr + 1;
+	minor = strtol(cp, &endptr, 10);
+	if (cp == endptr || *endptr != '.')
+		return 0;
+	cp = endptr + 1;
+	rev = strtol(cp, &endptr, 10);
+	if (cp == endptr)
+		return 0;
+	return ((((major * 256) + minor) * 256) + rev);
+}
+
+
+
 /*
  * This function sets the default parameters for a filesystem
  *
@@ -140,8 +166,8 @@ struct mke2fs_defaults {
 	{ default_str, 3, 1024, 8192 },
 	{ "journal", 0, 4096, 8192 },
 	{ "news", 0, 4096, 4096 },
-	{ "largefile", 0, DEF_MAX_BLOCKSIZE, 1024 * 1024 },
-	{ "largefile4", 0, DEF_MAX_BLOCKSIZE, 4096 * 1024 },
+	{ "largefile", 0, 4096, 1024 * 1024 },
+	{ "largefile4", 0, 4096, 4096 * 1024 },
 	{ 0, 0, 0, 0},
 };
 
@@ -171,11 +197,15 @@ static void set_fs_defaults(const char *fs_type,
 				blocksize : p->inode_ratio;
 		use_bsize = p->blocksize;
 	}
-	if (sector_size && use_bsize < sector_size)
-		use_bsize = sector_size;
 	if (blocksize <= 0) {
-		if (use_bsize == DEF_MAX_BLOCKSIZE)
+		if (use_bsize == DEF_MAX_BLOCKSIZE) {
 			use_bsize = sys_page_size;
+			if ((linux_version_code < (2*65536 + 6*256)) &&
+			    (use_bsize > 4096))
+				use_bsize = 4096;
+		}
+		if (sector_size && use_bsize < sector_size)
+			use_bsize = sector_size;
 		if ((blocksize < 0) && (use_bsize < (-blocksize)))
 			use_bsize = -blocksize;
 		blocksize = use_bsize;
@@ -189,7 +219,7 @@ static void set_fs_defaults(const char *fs_type,
 /*
  * Helper function for read_bb_file and test_disk
  */
-static void invalid_block(ext2_filsys fs, blk_t blk)
+static void invalid_block(ext2_filsys fs EXT2FS_ATTR((unused)), blk_t blk)
 {
 	fprintf(stderr, _("Bad block %u out of range; ignored.\n"), blk);
 	return;
@@ -250,8 +280,9 @@ static void test_disk(ext2_filsys fs, badblocks_list *bb_list)
 
 static void handle_bad_blocks(ext2_filsys fs, badblocks_list bb_list)
 {
-	int			i, j;
-	int			must_be_good;
+	dgrp_t			i;
+	blk_t			j;
+	unsigned 		must_be_good;
 	blk_t			blk;
 	badblocks_iterate	bb_iter;
 	errcode_t		retval;
@@ -274,7 +305,7 @@ static void handle_bad_blocks(ext2_filsys fs, badblocks_list bb_list)
 			fprintf(stderr, _("Blocks %d through %d must be good "
 				"in order to build a filesystem.\n"),
 				fs->super->s_first_data_block, must_be_good);
-			fprintf(stderr, _("Aborting....\n"));
+			fputs(_("Aborting....\n"), stderr);
 			exit(1);
 		}
 	}
@@ -327,6 +358,7 @@ struct progress_struct {
 	char		format[20];
 	char		backup[80];
 	__u32		max;
+	int		skip_progress;
 };
 
 static void progress_init(struct progress_struct *progress,
@@ -345,17 +377,21 @@ static void progress_init(struct progress_struct *progress,
 	sprintf(progress->format, "%%%dd/%%%dld", i, i);
 	memset(progress->backup, '\b', sizeof(progress->backup)-1);
 	progress->backup[sizeof(progress->backup)-1] = 0;
-	if ((2*i)+1 < sizeof(progress->backup))
+	if ((2*i)+1 < (int) sizeof(progress->backup))
 		progress->backup[(2*i)+1] = 0;
 	progress->max = max;
 
+	progress->skip_progress = 0;
+	if (getenv("MKE2FS_SKIP_PROGRESS"))
+		progress->skip_progress++;
+
 	fputs(label, stdout);
 	fflush(stdout);
 }
 
 static void progress_update(struct progress_struct *progress, __u32 val)
 {
-	if (progress->format[0] == 0)
+	if ((progress->format[0] == 0) || progress->skip_progress)
 		return;
 	printf(progress->format, val, progress->max);
 	fputs(progress->backup, stdout);
@@ -434,7 +470,8 @@ static void write_inode_tables(ext2_filsys fs)
 {
 	errcode_t	retval;
 	blk_t		blk;
-	int		i, num;
+	dgrp_t		i;
+	int		num;
 	struct progress_struct progress;
 
 	if (quiet)
@@ -487,7 +524,7 @@ static void create_root_dir(ext2_filsys fs)
 		inode.i_uid = getuid();
 		if (inode.i_uid)
 			inode.i_gid = getgid();
-		retval = ext2fs_write_inode(fs, EXT2_ROOT_INO, &inode);
+		retval = ext2fs_write_new_inode(fs, EXT2_ROOT_INO, &inode);
 		if (retval) {
 			com_err("ext2fs_write_inode", retval,
 				_("while setting root inode ownership"));
@@ -649,8 +686,10 @@ static void show_stats(ext2_filsys fs)
 {
 	struct ext2_super_block *s = fs->super;
 	char 			buf[80];
+        char                    *os;
 	blk_t			group_block;
-	int			i, need, col_left;
+	dgrp_t			i;
+	int			need, col_left;
 	
 	if (param.s_blocks_count != s->s_blocks_count)
 		fprintf(stderr, _("warning: %d blocks unused.\n\n"),
@@ -659,13 +698,10 @@ static void show_stats(ext2_filsys fs)
 	memset(buf, 0, sizeof(buf));
 	strncpy(buf, s->s_volume_name, sizeof(s->s_volume_name));
 	printf(_("Filesystem label=%s\n"), buf);
-	printf(_("OS type: "));
-	switch (fs->super->s_creator_os) {
-	    case EXT2_OS_LINUX: printf ("Linux"); break;
-	    case EXT2_OS_HURD:  printf ("GNU/Hurd");   break;
-	    case EXT2_OS_MASIX: printf ("Masix"); break;
-	    default:		printf (_("(unknown os)"));
-        }
+	fputs(_("OS type: "), stdout);
+        os = e2p_os2string(fs->super->s_creator_os);
+	fputs(os, stdout);
+	free(os);
 	printf("\n");
 	printf(_("Block size=%u (log=%u)\n"), fs->blocksize,
 		s->s_log_block_size);
@@ -677,6 +713,11 @@ static void show_stats(ext2_filsys fs)
 		s->s_r_blocks_count,
 	       100.0 * s->s_r_blocks_count / s->s_blocks_count);
 	printf(_("First data block=%u\n"), s->s_first_data_block);
+	if (s->s_reserved_gdt_blocks)
+		printf(_("Maximum filesystem blocks=%lu\n"),
+		       (s->s_reserved_gdt_blocks + fs->desc_blocks) *
+		       (fs->blocksize / sizeof(struct ext2_group_desc)) *
+		       s->s_blocks_per_group);
 	if (fs->group_desc_count > 1)
 		printf(_("%u block groups\n"), fs->group_desc_count);
 	else
@@ -689,7 +730,7 @@ static void show_stats(ext2_filsys fs)
 		printf("\n");
 		return;
 	}
-	
+
 	printf(_("Superblock backups stored on blocks: "));
 	group_block = s->s_first_data_block;
 	col_left = 0;
@@ -724,6 +765,10 @@ static int set_os(struct ext2_super_block *sb, char *os)
 		sb->s_creator_os = EXT2_OS_HURD;
 	else if (strcasecmp(os, "masix") == 0)
 		sb->s_creator_os = EXT2_OS_MASIX;
+	else if (strcasecmp(os, "freebsd") == 0)
+		sb->s_creator_os = EXT2_OS_FREEBSD;
+	else if (strcasecmp(os, "lites") == 0)
+		sb->s_creator_os = EXT2_OS_LITES;
 	else
 		return 0;
 	return 1;
@@ -731,17 +776,18 @@ static int set_os(struct ext2_super_block *sb, char *os)
 
 #define PATH_SET "PATH=/sbin"
 
-static void parse_raid_opts(const char *opts)
+static void parse_extended_opts(struct ext2_super_block *param, 
+				const char *opts)
 {
 	char	*buf, *token, *next, *p, *arg;
 	int	len;
-	int	raid_usage = 0;
+	int	r_usage = 0;
 
 	len = strlen(opts);
 	buf = malloc(len+1);
 	if (!buf) {
-		fprintf(stderr, _("Couldn't allocate memory to parse "
-			"raid options!\n"));
+		fprintf(stderr,
+			_("Couldn't allocate memory to parse options!\n"));
 		exit(1);
 	}
 	strcpy(buf, opts);
@@ -751,7 +797,7 @@ static void parse_raid_opts(const char *opts)
 		if (p) {
 			*p = 0;
 			next = p+1;
-		} 
+		}
 		arg = strchr(token, '=');
 		if (arg) {
 			*arg = 0;
@@ -759,32 +805,83 @@ static void parse_raid_opts(const char *opts)
 		}
 		if (strcmp(token, "stride") == 0) {
 			if (!arg) {
-				raid_usage++;
+				r_usage++;
 				continue;
 			}
 			fs_stride = strtoul(arg, &p, 0);
 			if (*p || (fs_stride == 0)) {
 				fprintf(stderr,
 					_("Invalid stride parameter.\n"));
-				raid_usage++;
+				r_usage++;
 				continue;
 			}
+		} else if (!strcmp(token, "resize")) {
+			unsigned long resize, bpg, rsv_groups;
+			unsigned long group_desc_count, desc_blocks;
+			unsigned int gdpb, blocksize;
+			int rsv_gdb;
+
+			if (!arg) {
+				r_usage++;
+				continue;
+			}
+
+			resize = parse_num_blocks(arg, 
+						  param->s_log_block_size);
+
+			if (resize == 0) {
+				fprintf(stderr, 
+					_("Invalid resize parameter: %s\n"),
+					arg);
+				r_usage++;
+				continue;
+			}
+			if (resize <= param->s_blocks_count) {
+				fprintf(stderr, 
+					_("The resize maximum must be greater than the filesystem size.\n"));
+				r_usage++;
+				continue;
+			}
+
+			blocksize = EXT2_BLOCK_SIZE(param);
+			bpg = param->s_blocks_per_group;
+			if (!bpg)
+				bpg = blocksize * 8;
+			gdpb = blocksize / sizeof(struct ext2_group_desc);
+			group_desc_count = (param->s_blocks_count +
+					    bpg - 1) / bpg;
+			desc_blocks = (group_desc_count +
+				       gdpb - 1) / gdpb;
+			rsv_groups = (resize + bpg - 1) / bpg;
+			rsv_gdb = (rsv_groups + gdpb - 1) / gdpb - 
+				desc_blocks;
+			if (rsv_gdb > EXT2_ADDR_PER_BLOCK(param))
+				rsv_gdb = EXT2_ADDR_PER_BLOCK(param);
+
+			if (rsv_gdb > 0) {
+				param->s_feature_compat |=
+					EXT2_FEATURE_COMPAT_RESIZE_INODE;
+
+				param->s_reserved_gdt_blocks = rsv_gdb;
+			}
 		} else
-			raid_usage++;
+			r_usage++;
 	}
-	if (raid_usage) {
-		fprintf(stderr, _("\nBad raid options specified.\n\n"
-			"Raid options are separated by commas, "
+	if (r_usage) {
+		fprintf(stderr, _("\nBad options specified.\n\n"
+			"Options are separated by commas, "
 			"and may take an argument which\n"
 			"\tis set off by an equals ('=') sign.\n\n"
 			"Valid raid options are:\n"
-			"\tstride=<stride length in blocks>\n\n"));
+			"\tstride=<stride length in blocks>\n"
+			"\tresize=<resize maximum size in blocks>\n\n"));
 		exit(1);
 	}
 }	
 
 static __u32 ok_features[3] = {
 	EXT3_FEATURE_COMPAT_HAS_JOURNAL |
+		EXT2_FEATURE_COMPAT_RESIZE_INODE |
 		EXT2_FEATURE_COMPAT_DIR_INDEX,	/* Compat */
 	EXT2_FEATURE_INCOMPAT_FILETYPE|		/* Incompat */
 		EXT3_FEATURE_INCOMPAT_JOURNAL_DEV|
@@ -803,12 +900,12 @@ static void PRS(int argc, char *argv[])
 	int		inode_size = 0;
 	int		reserved_ratio = 5;
 	int		sector_size = 0;
+	int		show_version_only = 0;
 	ext2_ino_t	num_inodes = 0;
 	errcode_t	retval;
 	char *		oldpath = getenv("PATH");
-	char *		raid_opts = 0;
+	char *		extended_opts = 0;
 	const char *	fs_type = 0;
-	int		default_features = 1;
 	blk_t		dev_size;
 #ifdef __linux__
 	struct 		utsname ut;
@@ -859,9 +956,8 @@ static void PRS(int argc, char *argv[])
 		perror("uname");
 		exit(1);
 	}
-	if ((ut.release[0] == '1') ||
-	    (ut.release[0] == '2' && ut.release[1] == '.' &&
-	     ut.release[2] < '2' && ut.release[3] == '.')) {
+	linux_version_code = parse_version_number(ut.release);
+	if (linux_version_code && linux_version_code < (2*65536 + 2*256)) {
 		param.s_rev_level = 0;
 		param.s_feature_incompat = 0;
 		param.s_feature_compat = 0;
@@ -878,7 +974,7 @@ static void PRS(int argc, char *argv[])
 	}
 
 	while ((c = getopt (argc, argv,
-		    "b:cf:g:i:jl:m:no:qr:R:s:tvI:J:ST:FL:M:N:O:V")) != EOF)
+		    "b:cE:f:g:i:jl:m:no:qr:R:s:tvI:J:ST:FL:M:N:O:V")) != EOF) {
 		switch (c) {
 		case 'b':
 			blocksize = strtol(optarg, &tmp, 0);
@@ -1019,14 +1115,12 @@ static void PRS(int argc, char *argv[])
 			mount_dir = optarg;
 			break;
 		case 'O':
-			if (!strcmp(optarg, "none") || default_features) {
+			if (!strcmp(optarg, "none")) {
 				param.s_feature_compat = 0;
 				param.s_feature_incompat = 0;
 				param.s_feature_ro_compat = 0;
-				default_features = 0;
-			}
-			if (!strcmp(optarg, "none"))
 				break;
+			}
 			if (e2p_edit_feature(optarg,
 					    &param.s_feature_compat,
 					    ok_features)) {
@@ -1035,8 +1129,9 @@ static void PRS(int argc, char *argv[])
 				exit(1);
 			}
 			break;
+		case 'E':
 		case 'R':
-			raid_opts = optarg;
+			extended_opts = optarg;
 			break;
 		case 'S':
 			super_only = 1;
@@ -1046,35 +1141,25 @@ static void PRS(int argc, char *argv[])
 			break;
 		case 'V':
 			/* Print version number and exit */
-			fprintf(stderr, _("\tUsing %s\n"),
-				error_message(EXT2_ET_BASE));
-			exit(0);
+			show_version_only++;
+			break;
 		default:
 			usage();
 		}
-	if (optind == argc)
-		usage();
-	device_name = argv[optind];
-	optind++;
-	if (optind < argc) {
-		unsigned long tmp2  = strtoul(argv[optind++], &tmp, 0);
-
-		if ((*tmp) || (tmp2 > 0xfffffffful)) {
-			com_err(program_name, 0, _("bad blocks count - %s"),
-				argv[optind - 1]);
-			exit(1);
-		}
-		param.s_blocks_count = tmp2;
 	}
-	if (optind < argc)
+	if ((optind == argc) && !show_version_only)
 		usage();
+	device_name = argv[optind++];
 
-	if (!quiet)
+	if (!quiet || show_version_only)
 		fprintf (stderr, "mke2fs %s (%s)\n", E2FSPROGS_VERSION, 
 			 E2FSPROGS_DATE);
 
-	if (raid_opts)
-		parse_raid_opts(raid_opts);
+	if (show_version_only) {
+		fprintf(stderr, _("\tUsing %s\n"), 
+			error_message(EXT2_ET_BASE));
+		exit(0);
+	}
 
 	/*
 	 * If there's no blocksize specified and there is a journal
@@ -1099,7 +1184,7 @@ static void PRS(int argc, char *argv[])
 				journal_device);
 			exit(1);
 		}
-		if ((blocksize < 0) && (jfs->blocksize < -blocksize)) {
+		if ((blocksize < 0) && (jfs->blocksize < (unsigned) (-blocksize))) {
 			com_err(program_name, 0,
 				_("Journal dev blocksize (%d) smaller than "
 				  "minimum blocksize %d\n"), jfs->blocksize,
@@ -1129,6 +1214,18 @@ static void PRS(int argc, char *argv[])
 			"blocksizes greater than 4096 \n\tusing ext3."
 			"  Use -b 4096 if this is an issue for you.\n\n");
 
+	if (optind < argc) {
+		param.s_blocks_count = parse_num_blocks(argv[optind++], 
+				param.s_log_block_size);
+		if (!param.s_blocks_count) {
+			com_err(program_name, 0, _("bad blocks count - %s"),
+				argv[optind - 1]);
+			exit(1);
+		}
+	}
+	if (optind < argc)
+		usage();
+
 	if (param.s_feature_incompat & EXT3_FEATURE_INCOMPAT_JOURNAL_DEV) {
 		if (!fs_type)
 			fs_type = "journal";
@@ -1151,10 +1248,20 @@ static void PRS(int argc, char *argv[])
 	if (noaction && param.s_blocks_count) {
 		dev_size = param.s_blocks_count;
 		retval = 0;
-	} else
+	} else {
+	retry:
 		retval = ext2fs_get_device_size(device_name,
 						EXT2_BLOCK_SIZE(&param),
 						&dev_size);
+		if ((retval == EFBIG) &&
+		    (blocksize == 0) && 
+		    (param.s_log_block_size == 0)) {
+			param.s_log_block_size = 2;
+			blocksize = 4096;
+			goto retry;
+		}
+	}
+			
 	if (retval && (retval != EXT2_ET_UNIMPLEMENTED)) {
 		com_err(program_name, retval,
 			_("while trying to determine filesystem size"));
@@ -1213,13 +1320,30 @@ static void PRS(int argc, char *argv[])
 			_("while trying to determine hardware sector size"));
 		exit(1);
 	}
+
+	if ((tmp = getenv("MKE2FS_DEVICE_SECTSIZE")) != NULL)
+		sector_size = atoi(tmp);
 	
 	set_fs_defaults(fs_type, &param, blocksize, sector_size, &inode_ratio);
 	blocksize = EXT2_BLOCK_SIZE(&param);
 	
+	if (extended_opts)
+		parse_extended_opts(&param, extended_opts);
+
+	/* Since sparse_super is the default, we would only have a problem
+	 * here if it was explicitly disabled.
+	 */
+	if ((param.s_feature_compat & EXT2_FEATURE_COMPAT_RESIZE_INODE) &&
+	    !(param.s_feature_ro_compat&EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER)) {
+		com_err(program_name, 0,
+			_("reserved online resize blocks not supported "
+			  "on non-sparse filesystem"));
+		exit(1);
+	}
+
 	if (param.s_blocks_per_group) {
 		if (param.s_blocks_per_group < 256 ||
-		    param.s_blocks_per_group > 8 * blocksize) {
+		    param.s_blocks_per_group > 8 * (unsigned) blocksize) {
 			com_err(program_name, 0,
 				_("blocks per group count out of range"));
 			exit(1);
@@ -1254,16 +1378,16 @@ static void PRS(int argc, char *argv[])
 	 * Calculate number of blocks to reserve
 	 */
 	param.s_r_blocks_count = (param.s_blocks_count * reserved_ratio) / 100;
-
 }
-					
+
 int main (int argc, char *argv[])
 {
 	errcode_t	retval = 0;
 	ext2_filsys	fs;
 	badblocks_list	bb_list = 0;
 	int		journal_blocks;
-	int		i, val;
+	unsigned int	i;
+	int		val;
 	io_manager	io_ptr;
 
 #ifdef ENABLE_NLS
@@ -1383,7 +1507,7 @@ int main (int argc, char *argv[])
 		fs->flags &= ~(EXT2_FLAG_IB_DIRTY|EXT2_FLAG_BB_DIRTY);
 	} else {
 		/* rsv must be a power of two (64kB is MD RAID sb alignment) */
-		int rsv = 65536 / fs->blocksize;
+		unsigned int rsv = 65536 / fs->blocksize;
 		unsigned long blocks = fs->super->s_blocks_count;
 		unsigned long start;
 		blk_t ret_blk;
@@ -1414,6 +1538,15 @@ int main (int argc, char *argv[])
 		create_lost_and_found(fs);
 		reserve_inodes(fs);
 		create_bad_block_inode(fs, bb_list);
+		if (fs->super->s_feature_compat & 
+		    EXT2_FEATURE_COMPAT_RESIZE_INODE) {
+			retval = ext2fs_create_resize_inode(fs);
+			if (retval) {
+				com_err("ext2fs_create_resize_inode", retval,
+				_("while reserving blocks for online resize"));
+				exit(1);
+			}
+		}
 	}
 
 	if (journal_device) {
@@ -1483,7 +1616,8 @@ no_journal:
 	}
 	if (!quiet) {
 		printf(_("done\n\n"));
-		print_check_message(fs);
+		if (!getenv("MKE2FS_SKIP_CHECK_MSG"))
+			print_check_message(fs);
 	}
 	val = ext2fs_close(fs);
 	return (retval || val) ? 1 : 0;
