@@ -25,6 +25,8 @@
  * 94/03/06	- Added the checks interval from Uwe Ohse (uwe@tirka.gun.de)
  */
 
+#define _XOPEN_SOURCE 500 /* for inclusion of strptime() */
+#define _BSD_SOURCE /* for inclusion of strcasecmp() */
 #include <fcntl.h>
 #include <grp.h>
 #ifdef HAVE_GETOPT_H
@@ -56,6 +58,7 @@ extern int optind;
 const char * program_name = "tune2fs";
 char * device_name;
 char * new_label, *new_last_mounted, *new_UUID;
+char * io_options;
 static int c_flag, C_flag, e_flag, f_flag, g_flag, i_flag, l_flag, L_flag;
 static int m_flag, M_flag, r_flag, s_flag = -1, u_flag, U_flag, T_flag;
 static time_t last_check_time;
@@ -154,7 +157,7 @@ static void remove_journal_device(ext2_filsys fs)
 	jsb = (journal_superblock_t *) buf;
 	if ((jsb->s_header.h_magic != (unsigned) ntohl(JFS_MAGIC_NUMBER)) ||
 	    (jsb->s_header.h_blocktype != (unsigned) ntohl(JFS_SUPERBLOCK_V2))) {
-		fprintf(stderr, _("Journal superblock not found!\n"));
+		fputs(_("Journal superblock not found!\n"), stderr);
 		goto no_valid_journal;
 	}
 
@@ -166,8 +169,8 @@ static void remove_journal_device(ext2_filsys fs)
 			break;
 	}
 	if (i >= nr_users) {
-		fprintf(stderr,
-			_("Filesystem's UUID not found on journal device.\n"));
+		fputs(_("Filesystem's UUID not found on journal device.\n"), 
+		      stderr);
 		commit_remove_journal = 1;
 		goto no_valid_journal;
 	}
@@ -187,19 +190,20 @@ static void remove_journal_device(ext2_filsys fs)
 
 no_valid_journal:
 	if (commit_remove_journal == 0) {
-		fprintf(stderr, _("Journal NOT removed\n"));
+		fputs(_("Journal NOT removed\n"), stderr);
 		exit(1);
 	}
 	fs->super->s_journal_dev = 0;
 	uuid_clear(fs->super->s_journal_uuid);
 	ext2fs_mark_super_dirty(fs);
-	printf(_("Journal removed\n"));
+	fputs(_("Journal removed\n"), stdout);
 	free(journal_path);
 }
 
 /* Helper function for remove_journal_inode */
 static int release_blocks_proc(ext2_filsys fs, blk_t *blocknr,
-			       int blockcnt, void *private)
+			       int blockcnt EXT2FS_ATTR((unused)), 
+			       void *private EXT2FS_ATTR((unused)))
 {
 	blk_t	block;
 	int	group;
@@ -279,6 +283,11 @@ static void update_feature_set(ext2_filsys fs, char *features)
 	int sparse, old_sparse, filetype, old_filetype;
 	int journal, old_journal, dxdir, old_dxdir;
 	struct ext2_super_block *sb= fs->super;
+	__u32	old_compat, old_incompat, old_ro_compat;
+
+	old_compat = sb->s_feature_compat;
+	old_ro_compat = sb->s_feature_ro_compat;
+	old_incompat = sb->s_feature_incompat;
 
 	old_sparse = sb->s_feature_ro_compat &
 		EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER;
@@ -305,19 +314,17 @@ static void update_feature_set(ext2_filsys fs, char *features)
 	if (old_journal && !journal) {
 		if ((mount_flags & EXT2_MF_MOUNTED) &&
 		    !(mount_flags & EXT2_MF_READONLY)) {
-			fprintf(stderr,
-				_("The has_journal flag may only be "
-				  "cleared when the filesystem is\n"
-				  "unmounted or mounted "
-				  "read-only.\n"));
+			fputs(_("The has_journal flag may only be "
+				"cleared when the filesystem is\n"
+				"unmounted or mounted "
+				"read-only.\n"), stderr);
 			exit(1);
 		}
 		if (sb->s_feature_incompat &
 		    EXT3_FEATURE_INCOMPAT_RECOVER) {
-			fprintf(stderr,
-				_("The needs_recovery flag is set.  "
-				  "Please run e2fsck before clearing\n"
-				  "the has_journal flag.\n"));
+			fputs(_("The needs_recovery flag is set.  "
+				"Please run e2fsck before clearing\n"
+				"the has_journal flag.\n"), stderr);
 			exit(1);
 		}
 		if (sb->s_journal_inum) {
@@ -354,7 +361,10 @@ static void update_feature_set(ext2_filsys fs, char *features)
 		sb->s_state &= ~EXT2_VALID_FS;
 		printf("\n%s\n", _(please_fsck));
 	}
-	ext2fs_mark_super_dirty(fs);
+	if ((old_compat != sb->s_feature_compat) ||
+	    (old_ro_compat != sb->s_feature_ro_compat) ||
+	    (old_incompat != sb->s_feature_incompat))
+		ext2fs_mark_super_dirty(fs);
 }
 
 /*
@@ -369,7 +379,7 @@ static void add_journal(ext2_filsys fs)
 
 	if (fs->super->s_feature_compat &
 	    EXT3_FEATURE_COMPAT_HAS_JOURNAL) {
-		fprintf(stderr, _("The filesystem already has a journal.\n"));
+		fputs(_("The filesystem already has a journal.\n"), stderr);
 		goto err;
 	}
 	if (journal_device) {
@@ -402,9 +412,9 @@ static void add_journal(ext2_filsys fs)
 				 journal_device);
 			goto err;
 		}
-		printf(_("done\n"));
+		fputs(_("done\n"), stdout);
 	} else if (journal_size) {
-		printf(_("Creating journal inode: "));
+		fputs(_("Creating journal inode: "), stdout);
 		fflush(stdout);
 		journal_blocks = figure_journal_size(journal_size, fs);
 
@@ -416,7 +426,7 @@ static void add_journal(ext2_filsys fs)
 				_("\n\twhile trying to create journal file"));
 			exit(1);
 		} else
-			printf(_("done\n"));
+			fputs(_("done\n"), stdout);
 		/*
 		 * If the filesystem wasn't mounted, we need to force
 		 * the block group descriptors out.
@@ -437,10 +447,18 @@ err:
 static void parse_e2label_options(int argc, char ** argv)
 {
 	if ((argc < 2) || (argc > 3)) {
-		fprintf(stderr, _("Usage: e2label device [newlabel]\n"));
+		fputs(_("Usage: e2label device [newlabel]\n"), stderr);
 		exit(1);
 	}
+	io_options = strchr(argv[1], '?');
+	if (io_options)
+		*io_options++ = 0;
 	device_name = blkid_get_devname(NULL, argv[1], NULL);
+	if (!device_name) {
+		com_err("e2label", 0, _("Unable to resolve '%s'"), 
+			argv[1]);
+		exit(1);
+	}
 	if (argc == 3) {
 		open_flag = EXT2_FLAG_RW | EXT2_FLAG_JOURNAL_DEV_OK;
 		L_flag = 1;
@@ -458,9 +476,9 @@ static time_t parse_time(char *str)
 	}
 	memset(&ts, 0, sizeof(ts));
 #ifdef HAVE_STRPTIME
-	strptime(optarg, "%Y%m%d%H%M%S", &ts);
+	strptime(str, "%Y%m%d%H%M%S", &ts);
 #else
-	sscanf(optarg, "%4d%2d%2d%2d%2d%2d", &ts.tm_year, &ts.tm_mon,
+	sscanf(str, "%4d%2d%2d%2d%2d%2d", &ts.tm_year, &ts.tm_mon,
 	       &ts.tm_mday, &ts.tm_hour, &ts.tm_min, &ts.tm_sec);
 	ts.tm_year -= 1900;
 	ts.tm_mon -= 1;
@@ -690,7 +708,15 @@ static void parse_tune2fs_options(int argc, char **argv)
 		usage();
 	if (!open_flag && !l_flag)
 		usage();
+	io_options = strchr(argv[optind], '?');
+	if (io_options)
+		*io_options++ = 0;
 	device_name = blkid_get_devname(NULL, argv[optind], NULL);
+	if (!device_name) {
+		com_err("tune2fs", 0, _("Unable to resolve '%s'"), 
+			argv[optind]);
+		exit(1);
+	}
 }
 
 void do_findfs(int argc, char **argv)
@@ -704,7 +730,7 @@ void do_findfs(int argc, char **argv)
 	}
 	dev = blkid_get_devname(NULL, argv[1], NULL);
 	if (!dev) {
-		fprintf(stderr, "Filesystem matching %s not found\n", 
+		com_err("findfs", 0, _("Unable to resolve '%s'"), 
 			argv[1]);
 		exit(1);
 	}
@@ -743,7 +769,8 @@ int main (int argc, char ** argv)
 #else
 	io_ptr = unix_io_manager;
 #endif
-	retval = ext2fs_open (device_name, open_flag, 0, 0, io_ptr, &fs);
+	retval = ext2fs_open2(device_name, io_options, open_flag, 
+			      0, 0, io_ptr, &fs);
         if (retval) {
 		com_err (program_name, retval, _("while trying to open %s"),
 			 device_name);
@@ -802,9 +829,9 @@ int main (int argc, char ** argv)
 			reserved_ratio, sb->s_r_blocks_count);
 	}
 	if (r_flag) {
-		if (reserved_blocks >= sb->s_blocks_count) {
+		if (reserved_blocks >= sb->s_blocks_count/2) {
 			com_err (program_name, 0,
-				 _("reserved blocks count is too big (%ul)"),
+				 _("reserved blocks count is too big (%lu)"),
 				 reserved_blocks);
 			exit (1);
 		}
@@ -816,8 +843,8 @@ int main (int argc, char ** argv)
 	if (s_flag == 1) {
 		if (sb->s_feature_ro_compat &
 		    EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER)
-			fprintf(stderr, _("\nThe filesystem already"
-				" has sparse superblocks.\n"));
+			fputs(_("\nThe filesystem already has sparse "
+				"superblocks.\n"), stderr);
 		else {
 			sb->s_feature_ro_compat |=
 				EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER;
@@ -830,8 +857,8 @@ int main (int argc, char ** argv)
 	if (s_flag == 0) {
 		if (!(sb->s_feature_ro_compat &
 		      EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER))
-			fprintf(stderr, _("\nThe filesystem already"
-				" has sparse superblocks disabled.\n"));
+			fputs(_("\nThe filesystem already has sparse "
+				"superblocks disabled.\n"), stderr);
 		else {
 			sb->s_feature_ro_compat &=
 				~EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER;
@@ -855,8 +882,8 @@ int main (int argc, char ** argv)
 	}
 	if (L_flag) {
 		if (strlen(new_label) > sizeof(sb->s_volume_name))
-			fprintf(stderr, _("Warning: label too "
-				"long, truncating.\n"));
+			fputs(_("Warning: label too long, truncating.\n"), 
+			      stderr);
 		memset(sb->s_volume_name, 0, sizeof(sb->s_volume_name));
 		strncpy(sb->s_volume_name, new_label,
 			sizeof(sb->s_volume_name));
