@@ -34,6 +34,7 @@
  */
 
 #include "resize2fs.h"
+#include <time.h>
 
 #ifdef __linux__			/* Kludge for debugging */
 #define RESIZE2FS_DEBUG
@@ -45,6 +46,7 @@ static errcode_t block_mover(ext2_resize_t rfs);
 static errcode_t inode_scan_and_fix(ext2_resize_t rfs);
 static errcode_t inode_ref_fix(ext2_resize_t rfs);
 static errcode_t move_itables(ext2_resize_t rfs);
+static errcode_t fix_resize_inode(ext2_filsys fs);
 static errcode_t ext2fs_calculate_summary_stats(ext2_filsys fs);
 
 /*
@@ -81,8 +83,7 @@ errcode_t resize_fs(ext2_filsys fs, blk_t *new_size, int flags,
 	/*
 	 * Create the data structure
 	 */
-	retval = ext2fs_get_mem(sizeof(struct ext2_resize_struct),
-				(void **) &rfs);
+	retval = ext2fs_get_mem(sizeof(struct ext2_resize_struct), &rfs);
 	if (retval)
 		return retval;
 	memset(rfs, 0, sizeof(struct ext2_resize_struct));
@@ -125,11 +126,15 @@ errcode_t resize_fs(ext2_filsys fs, blk_t *new_size, int flags,
 	if (retval)
 		goto errout;
 
+	retval = move_itables(rfs);
+	if (retval)
+		goto errout;
+
 	retval = ext2fs_calculate_summary_stats(rfs->new_fs);
 	if (retval)
 		goto errout;
 	
-	retval = move_itables(rfs);
+	retval = fix_resize_inode(rfs->new_fs);
 	if (retval)
 		goto errout;
 
@@ -141,8 +146,8 @@ errcode_t resize_fs(ext2_filsys fs, blk_t *new_size, int flags,
 	
 	ext2fs_free(rfs->old_fs);
 	if (rfs->itable_buf)
-		ext2fs_free_mem((void **) &rfs->itable_buf);
-	ext2fs_free_mem((void **) &rfs);
+		ext2fs_free_mem(&rfs->itable_buf);
+	ext2fs_free_mem(&rfs);
 	
 	return 0;
 
@@ -150,8 +155,8 @@ errout:
 	if (rfs->new_fs)
 		ext2fs_free(rfs->new_fs);
 	if (rfs->itable_buf)
-		ext2fs_free_mem((void **) &rfs->itable_buf);
-	ext2fs_free_mem((void **) &rfs);
+		ext2fs_free_mem(&rfs->itable_buf);
+	ext2fs_free_mem(&rfs);
 	return retval;
 }
 
@@ -177,9 +182,10 @@ static errcode_t adjust_superblock(ext2_resize_t rfs, blk_t new_size)
 	errcode_t	retval;
 	ext2_ino_t	real_end;
 	blk_t		blk, group_block;
-	unsigned long	i, j;
+	unsigned long	i, j, old_desc_blocks;
 	int		old_numblocks, numblocks, adjblocks;
-	int		has_super, meta_bg, meta_bg_size, old_desc_blocks;
+	unsigned int	meta_bg, meta_bg_size;
+	int		has_super;
 	unsigned long	max_group;
 	
 	fs = rfs->new_fs;
@@ -204,12 +210,13 @@ retry:
 	 * includes the superblock backup, the group descriptor
 	 * backups, the inode bitmap, the block bitmap, and the inode
 	 * table.
-	 *
-	 * XXX Not all block groups need the descriptor blocks, but
-	 * being clever is tricky...
 	 */
-	overhead = 3 + fs->desc_blocks + fs->inode_blocks_per_group;
-	
+	overhead = (int) (2 + fs->inode_blocks_per_group);
+
+	if (ext2fs_bg_has_super(fs, fs->group_desc_count - 1))
+		overhead += 1 + fs->desc_blocks + 
+			fs->super->s_reserved_gdt_blocks;
+
 	/*
 	 * See if the last group is big enough to support the
 	 * necessary data structures.  If not, we need to get rid of
@@ -271,7 +278,7 @@ retry:
 		retval = ext2fs_resize_mem(rfs->old_fs->desc_blocks *
 					   fs->blocksize,
 					   fs->desc_blocks * fs->blocksize,
-					   (void **) &fs->group_desc);
+					   &fs->group_desc);
 		if (retval)
 			goto errout;
 	}
@@ -286,6 +293,29 @@ retry:
 		goto errout;
 	}
 
+	/*
+	 * If the resize_inode feature is set, and we are changing the
+	 * number of descriptor blocks, then adjust
+	 * s_reserved_gdt_blocks if possible to avoid needing to move
+	 * the inode table either now or in the future.
+	 */
+	if ((fs->super->s_feature_compat & 
+	     EXT2_FEATURE_COMPAT_RESIZE_INODE) &&
+	    (rfs->old_fs->desc_blocks != fs->desc_blocks)) {
+		int new;
+
+		new = ((int) fs->super->s_reserved_gdt_blocks) + 
+			(rfs->old_fs->desc_blocks - fs->desc_blocks);
+		if (new < 0)
+			new = 0;
+		if (new > fs->blocksize/4)
+			new = fs->blocksize/4;
+		fs->super->s_reserved_gdt_blocks = new;
+		if (new == 0)
+			fs->super->s_feature_compat &= 
+				~EXT2_FEATURE_COMPAT_RESIZE_INODE;
+	}
+
 	/*
 	 * If we are shrinking the number block groups, we're done and
 	 * can exit now.
@@ -326,7 +356,7 @@ retry:
 	 * Initialize the new block group descriptors
 	 */
 	retval = ext2fs_get_mem(fs->blocksize * fs->inode_blocks_per_group,
-				(void **) &rfs->itable_buf);
+				&rfs->itable_buf);
 	if (retval)
 		goto errout;
 
@@ -345,7 +375,8 @@ retry:
 	if (fs->super->s_feature_incompat & EXT2_FEATURE_INCOMPAT_META_BG)
 		old_desc_blocks = fs->super->s_first_meta_bg;
 	else
-		old_desc_blocks = fs->desc_blocks;
+		old_desc_blocks = fs->desc_blocks + 
+			fs->super->s_reserved_gdt_blocks;
 	for (i = rfs->old_fs->group_desc_count;
 	     i < fs->group_desc_count; i++) {
 		memset(&fs->group_desc[i], 0,
@@ -444,61 +475,31 @@ errout:
  * filesystem meta-data blocks.
  */
 static errcode_t mark_table_blocks(ext2_filsys fs,
-				   ext2fs_block_bitmap *ret_bmap)
+				   ext2fs_block_bitmap bmap)
 {
 	blk_t			block, b;
-	int			i,j, has_super, meta_bg, meta_bg_size;
-	int			old_desc_blocks;
-	ext2fs_block_bitmap	bmap;
+	unsigned int		j;
+	dgrp_t			i;
+	unsigned long		meta_bg, meta_bg_size;
+	int			has_super;
+	unsigned int		old_desc_blocks;
 	errcode_t		retval;
 
-	retval = ext2fs_allocate_block_bitmap(fs, _("meta-data blocks"), 
-					      &bmap);
-	if (retval)
-		return retval;
-	
 	meta_bg_size = (fs->blocksize / sizeof (struct ext2_group_desc));
 	block = fs->super->s_first_data_block;
 	if (fs->super->s_feature_incompat & EXT2_FEATURE_INCOMPAT_META_BG)
 		old_desc_blocks = fs->super->s_first_meta_bg;
 	else
-		old_desc_blocks = fs->desc_blocks;
+		old_desc_blocks = fs->desc_blocks + 
+			fs->super->s_reserved_gdt_blocks;
 	for (i = 0; i < fs->group_desc_count; i++) {
-		has_super = ext2fs_bg_has_super(fs, i);
-		if (has_super)
-			/*
-			 * Mark this group's copy of the superblock
-			 */
-			ext2fs_mark_block_bitmap(bmap, block);
-		
-		meta_bg = i / meta_bg_size;
-		
-		if (!(fs->super->s_feature_incompat &
-		      EXT2_FEATURE_INCOMPAT_META_BG) ||
-		    (meta_bg < fs->super->s_first_meta_bg)) {
-			if (has_super) {
-				/*
-				 * Mark this group's copy of the descriptors
-				 */
-				for (j = 0; j < old_desc_blocks; j++)
-					ext2fs_mark_block_bitmap(bmap,
-							 block + j + 1);
-			}
-		} else {
-			if (has_super)
-				has_super = 1;
-			if (((i % meta_bg_size) == 0) ||
-			    ((i % meta_bg_size) == 1) ||
-			    ((i % meta_bg_size) == (meta_bg_size-1)))
-				ext2fs_mark_block_bitmap(bmap,
-							 block + has_super);
-		}
+		ext2fs_reserve_super_and_bgd(fs, i, bmap);
 	
 		/*
 		 * Mark the blocks used for the inode table
 		 */
 		for (j = 0, b = fs->group_desc[i].bg_inode_table;
-		     j < fs->inode_blocks_per_group;
+		     j < (unsigned int) fs->inode_blocks_per_group;
 		     j++, b++)
 			ext2fs_mark_block_bitmap(bmap, b);
 			    
@@ -507,6 +508,7 @@ static errcode_t mark_table_blocks(ext2_filsys fs,
 		 */
 		ext2fs_mark_block_bitmap(bmap,
 					 fs->group_desc[i].bg_block_bitmap);
+
 		/*
 		 * Mark block used for the inode bitmap 
 		 */
@@ -514,7 +516,6 @@ static errcode_t mark_table_blocks(ext2_filsys fs,
 					 fs->group_desc[i].bg_inode_bitmap);
 		block += fs->super->s_blocks_per_group;
 	}
-	*ret_bmap = bmap;
 	return 0;
 }
 
@@ -561,9 +562,11 @@ static void mark_fs_metablock(ext2_resize_t rfs,
  */
 static errcode_t blocks_to_move(ext2_resize_t rfs)
 {
-	int	i, j, max_groups, has_super, meta_bg, meta_bg_size;
-	blk_t	blk, group_blk;
-	unsigned long old_blocks, new_blocks;
+	int		j, has_super;
+	dgrp_t		i, max_groups;
+	blk_t		blk, group_blk;
+	unsigned long	old_blocks, new_blocks;
+	unsigned int	meta_bg, meta_bg_size;
 	errcode_t	retval;
 	ext2_filsys 	fs, old_fs;
 	ext2fs_block_bitmap	meta_bmap;
@@ -583,7 +586,12 @@ static errcode_t blocks_to_move(ext2_resize_t rfs)
 	if (retval)
 		return retval;
 
-	retval = mark_table_blocks(old_fs, &meta_bmap);
+	retval = ext2fs_allocate_block_bitmap(fs, _("meta-data blocks"), 
+					      &meta_bmap);
+	if (retval)
+		return retval;
+	
+	retval = mark_table_blocks(old_fs, meta_bmap);
 	if (retval)
 		return retval;
 
@@ -607,8 +615,8 @@ static errcode_t blocks_to_move(ext2_resize_t rfs)
 		old_blocks = old_fs->super->s_first_meta_bg;
 		new_blocks = fs->super->s_first_meta_bg;
 	} else {
-		old_blocks = old_fs->desc_blocks;
-		new_blocks = fs->desc_blocks;
+		old_blocks = old_fs->desc_blocks + old_fs->super->s_reserved_gdt_blocks;
+		new_blocks = fs->desc_blocks + fs->super->s_reserved_gdt_blocks;
 	}
 	
 	if (old_blocks == new_blocks) {
@@ -848,7 +856,7 @@ static errcode_t block_mover(ext2_resize_t rfs)
 	if (!rfs->itable_buf) {
 		retval = ext2fs_get_mem(fs->blocksize *
 					fs->inode_blocks_per_group,
-					(void **) &rfs->itable_buf);
+					&rfs->itable_buf);
 		if (retval)
 			return retval;
 	}
@@ -970,8 +978,9 @@ struct process_block_struct {
 };
 
 static int process_block(ext2_filsys fs, blk_t	*block_nr,
-			 e2_blkcnt_t blockcnt, blk_t ref_block,
-			 int ref_offset, void *priv_data)
+			 e2_blkcnt_t blockcnt, 
+			 blk_t ref_block EXT2FS_ATTR((unused)),
+			 int ref_offset EXT2FS_ATTR((unused)), void *priv_data)
 {
 	struct process_block_struct *pb;
 	errcode_t	retval;
@@ -1008,7 +1017,8 @@ static int process_block(ext2_filsys fs, blk_t	*block_nr,
 /*
  * Progress callback
  */
-static errcode_t progress_callback(ext2_filsys fs, ext2_inode_scan scan,
+static errcode_t progress_callback(ext2_filsys fs, 
+				   ext2_inode_scan scan EXT2FS_ATTR((unused)),
 				   dgrp_t group, void * priv_data)
 {
 	ext2_resize_t rfs = (ext2_resize_t) priv_data;
@@ -1065,8 +1075,7 @@ static errcode_t inode_scan_and_fix(ext2_resize_t rfs)
 
 	retval = ext2fs_init_dblist(rfs->old_fs, 0);
 	if (retval) goto errout;
-	retval = ext2fs_get_mem(rfs->old_fs->blocksize * 3,
-				(void **) &block_buf);
+	retval = ext2fs_get_mem(rfs->old_fs->blocksize * 3, &block_buf);
 	if (retval) goto errout;
 
 	start_to_move = (rfs->new_fs->group_desc_count *
@@ -1093,7 +1102,7 @@ static errcode_t inode_scan_and_fix(ext2_resize_t rfs)
 		if (!ino)
 			break;
 
-		if (inode.i_links_count == 0)
+		if (inode.i_links_count == 0 && ino != EXT2_RESIZE_INO)
 			continue; /* inode not in use */
 
 		pb.is_dir = LINUX_S_ISDIR(inode.i_mode);
@@ -1176,7 +1185,7 @@ errout:
 	if (scan)
 		ext2fs_close_inode_scan(scan);
 	if (block_buf)
-		ext2fs_free_mem((void **) &block_buf);
+		ext2fs_free_mem(&block_buf);
 	return retval;
 }
 
@@ -1194,9 +1203,12 @@ struct istruct {
 	int		num;
 };
 
-static int check_and_change_inodes(ext2_ino_t dir, int entry,
+static int check_and_change_inodes(ext2_ino_t dir, 
+				   int entry EXT2FS_ATTR((unused)),
 				   struct ext2_dir_entry *dirent, int offset,
-				   int	blocksize, char *buf, void *priv_data)
+				   int	blocksize EXT2FS_ATTR((unused)),
+				   char *buf EXT2FS_ATTR((unused)), 
+				   void *priv_data)
 {
 	struct istruct *is = (struct istruct *) priv_data;
 	struct ext2_inode 	inode;
@@ -1222,7 +1234,7 @@ static int check_and_change_inodes(ext2_ino_t dir, int entry,
 #ifdef RESIZE2FS_DEBUG
 	if (is->rfs->flags & RESIZE_DEBUG_INODEMAP)
 		printf(_("Inode translate (dir=%u, name=%.*s, %u->%u)\n"),
-		       dir, dirent->name_len, dirent->name,
+		       dir, dirent->name_len&0xFF, dirent->name,
 		       dirent->inode, new_inode);
 #endif
 
@@ -1301,12 +1313,13 @@ errout:
  */
 static errcode_t move_itables(ext2_resize_t rfs)
 {
-	int		i, n, num, max_groups, size, diff;
+	int		n, num, size, diff;
+	dgrp_t		i, max_groups;
 	ext2_filsys	fs = rfs->new_fs;
 	char		*cp;
-	blk_t		old_blk, new_blk;
+	blk_t		old_blk, new_blk, blk;
 	errcode_t	retval;
-	int		to_move, moved;
+	int		j, to_move, moved;
 
 	max_groups = fs->group_desc_count;
 	if (max_groups > rfs->old_fs->group_desc_count)
@@ -1314,7 +1327,7 @@ static errcode_t move_itables(ext2_resize_t rfs)
 
 	size = fs->blocksize * fs->inode_blocks_per_group;
 	if (!rfs->itable_buf) {
-		retval = ext2fs_get_mem(size, (void **) &rfs->itable_buf);
+		retval = ext2fs_get_mem(size, &rfs->itable_buf);
 		if (retval)
 			return retval;
 	}
@@ -1395,16 +1408,23 @@ static errcode_t move_itables(ext2_resize_t rfs)
 			if (retval)
 				goto errout;
 		}
+
+		for (blk = rfs->old_fs->group_desc[i].bg_inode_table, j=0;
+		     j < fs->inode_blocks_per_group ; j++, blk++)
+			ext2fs_unmark_block_bitmap(fs->block_map, blk);
+
 		rfs->old_fs->group_desc[i].bg_inode_table = new_blk;
 		ext2fs_mark_super_dirty(rfs->old_fs);
+		ext2fs_flush(rfs->old_fs);
+
 		if (rfs->progress) {
-			ext2fs_flush(rfs->old_fs);
 			retval = rfs->progress(rfs, E2_RSZ_MOVE_ITABLE_PASS,
 					       ++moved, to_move);
 			if (retval)
 				goto errout;
 		}
 	}
+	mark_table_blocks(fs, fs->block_map);
 	ext2fs_flush(fs);
 #ifdef RESIZE2FS_DEBUG
 	if (rfs->flags & RESIZE_DEBUG_ITABLEMOVE) 
@@ -1416,6 +1436,57 @@ errout:
 	return retval;
 }
 
+/*
+ * Fix the resize inode 
+ */
+static errcode_t fix_resize_inode(ext2_filsys fs)
+{
+	struct ext2_inode	inode;
+	errcode_t		retval;
+	char *			block_buf;
+
+	if (!(fs->super->s_feature_compat & 
+	      EXT2_FEATURE_COMPAT_RESIZE_INODE))
+		return 0;
+
+	retval = ext2fs_get_mem(fs->blocksize, &block_buf);
+	if (retval) goto errout;
+
+	retval = ext2fs_read_inode(fs, EXT2_RESIZE_INO, &inode);
+	if (retval) goto errout;
+
+	inode.i_blocks = fs->blocksize/512;
+
+	retval = ext2fs_write_inode(fs, EXT2_RESIZE_INO, &inode);
+	if (retval) goto errout;
+
+	if (!inode.i_block[EXT2_DIND_BLOCK]) {
+		/* 
+		 * Avoid zeroing out block #0; that's rude.  This
+		 * should never happen anyway since the filesystem
+		 * should be fsck'ed and we assume it is consistent.
+		 */
+		fprintf(stderr, 
+			_("Should never happen resize inode corrupt!\n"));
+		exit(1);
+	}
+
+	memset(block_buf, 0, fs->blocksize);
+
+	retval = io_channel_write_blk(fs->io, inode.i_block[EXT2_DIND_BLOCK],
+				      1, block_buf);
+	if (retval) goto errout;
+	
+	retval = ext2fs_create_resize_inode(fs);
+	if (retval)
+		goto errout;
+
+errout:
+	if (block_buf)
+		ext2fs_free_mem(&block_buf);
+	return retval;
+}
+
 /*
  * Finally, recalculate the summary information
  */
@@ -1423,8 +1494,8 @@ static errcode_t ext2fs_calculate_summary_stats(ext2_filsys fs)
 {
 	blk_t		blk;
 	ext2_ino_t	ino;
-	int		group = 0;
-	int		count = 0;
+	unsigned int	group = 0;
+	unsigned int	count = 0;
 	int		total_free = 0;
 	int		group_free = 0;
 
